# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: Remove stripe artefacts
   :platform: Unix
   :synopsis: A plugin working in sinogram space to remove stripe artefacts
.. moduleauthor:: Nghia Vo <scientificsoftware@diamond.ac.uk>

"""

from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin
from savu.data.plugin_list import CitationInformation

import numpy as np
from scipy.ndimage import median_filter
import pyfftw.interfaces.scipy_fftpack as fft
from scipy.signal import savgol_filter

@register_plugin
class RingRemovalFitting(Plugin, CpuPlugin):
    """

    Method to remove stripe artefacts in a sinogram (<-> ring artefacts in a \
    reconstructed image) using a fitting-based method. It's suitable for use \
    on low-contrast sinogram and/or for removing blurry rings.

    :param sigmax: Sigma of the Gaussian window in x-direction which controls\
     the strength of the removal. Default: 5.
    :param sigmay: Sigma of the Gaussian window in y-direction. Default: 20.
    :param order: polynomial fit order. Default: 2.
    """

    def __init__(self):
        super(RingRemovalFitting, self).__init__(
                "RingRemovalFitting")

    def setup(self):
        in_dataset, out_dataset = self.get_datasets()
        out_dataset[0].create_dataset(in_dataset[0])
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('SINOGRAM', 'single')
        out_pData[0].plugin_data_setup('SINOGRAM', 'single')

    def _create_2d_window(self, height, width, sigmax, sigmay):
        """
        Create a 2D Gaussian window.
        ---------
        Parameters: - height, width: shape of the window.
                    - sigmax, sigmay: sigmas of the window.
        ---------
        Return:     - 2D window.
        """
        centerx = (width-1.0) / 2.0
        centery = (height-1.0) / 2.0
        y, x = np.ogrid[-centery:height-centery, -centerx:width-centerx]
        numx = 2.0*sigmax*sigmax
        numy = 2.0*sigmay*sigmay
        win2d = np.exp(-(x*x/numx+y*y/numy))
        return win2d

    def pre_process(self):
        in_pData = self.get_plugin_in_datasets()
        width_dim = \
            in_pData[0].get_data_dimension_by_axis_label('detector_x')
        height_dim = \
            in_pData[0].get_data_dimension_by_axis_label('rotation_angle')
        sino_shape = list(in_pData[0].get_shape())
        self.pad = 50 # To reduce artifact caused by FFT
        self.width1 = sino_shape[width_dim] + 2*self.pad
        self.height1 = sino_shape[height_dim] + 2*self.pad
        sigmax = np.clip(np.int16(
            self.parameters['sigmax']), 1, self.width1-1)
        sigmay = np.clip(np.int16(
            self.parameters['sigmay']), 1, self.height1-1)
        self.window2d = self._create_2d_window(
            self.height1, self.width1, sigmax, sigmay)
        self.order = np.clip(
            np.int16(self.parameters['order']), 0, self.height1-1)
        listx = np.arange(0, self.width1)
        listy = np.arange(0, self.height1)
        x,y = np.meshgrid(listx, listy)
        self.matsign = np.power(-1.0,x+y)

    def process_frames(self, data):
        sinogram = data[0]
        (height, _) = sinogram.shape
        if height%2==0:
            height = height - 1
        sinofit = np.abs(savgol_filter(
            sinogram, height, self.order, axis=0, mode = 'mirror'))
        sinofit2 = np.pad(
            sinofit,((0, 0), (self.pad, self.pad)), mode = 'edge')
        sinofit2 = np.pad(
            sinofit2,((self.pad, self.pad), (0, 0)), mode = 'mean')
        sinofitsmooth = np.real(fft.ifft2(fft.fft2(
            sinofit2*self.matsign)*self.window2d)*self.matsign)
        sinofitsmooth = sinofitsmooth[self.pad:self.height1-self.pad,
                                      self.pad:self.width1-self.pad]
        num1 = np.mean(sinofit)
        num2 = np.mean(sinofitsmooth)
        sinofitsmooth = num1*sinofitsmooth/num2
        return sinogram/sinofit*sinofitsmooth

    def get_citation_information(self):
        cite_info = CitationInformation()
        cite_info.description = \
            ("The code of ring removal is the implementation of the work of \
            Nghia T. Vo et al. taken from algorithm 1 in this paper.")
        cite_info.bibtex = \
            ("@article{Vo:18,\n" +
             "title={Superior techniques for eliminating ring artifacts in\
              X-ray micro-tomography},\n" +
             "author={Nghia T. Vo, Robert C. Atwood,\
              and Michael Drakopoulos},\n" +
             "journal={Opt. Express},\n" +
             "volume={26},\n" +
             "number={22},\n" +
             "pages={28396--28412},\n" +
             "year={2018},\n" +
             "publisher={OSA}" +
             "}")
        cite_info.doi = "doi: DOI: 10.1364/OE.26.028396"
        return cite_info



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
.. module:: iterative_plugin
   :platform: Unix
   :synopsis: Iterative plugin example
.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>
"""

import numpy as np

from savu.plugins.utils import register_plugin
from savu.plugins.filters.base_filter import BaseFilter
from savu.plugins.driver.iterative_plugin import IterativePlugin


@register_plugin
class TestingIterativePlugin(BaseFilter, IterativePlugin):
    """
    A plugin to test the iterative plugin driver
    
    :u*param nIterations: Number of iterations.  Default: 10.

    """

    def __init__(self):
        super(TestingIterativePlugin, self).__init__("TestingIterativePlugin")

    def pre_process(self):
        self.set_iterations(self.parameters['nIterations'])

    def process_frames(self, data):
        # A random example function
        if self.get_iteration() == 0:
            return np.zeros(data[0].shape, dtype=np.float32)
        return data[1] + np.ones(data[0].shape, dtype=np.float32)*10

    def post_process(self):
        # option here to break out of the iterations
        #self.set_processing_complete()
        pass

    def setup(self):
        # set up the output dataset that is created by the plugin
        in_dataset, out_dataset = self.get_datasets()

        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('SINOGRAM', 'single')

        # Cloned datasets are at the end of the out_dataset list
        out_dataset[0].create_dataset(in_dataset[0])

        # What is a cloned dataset?
        # Since each dataset in Savu has its own backing hdf5 file, a dataset
        # cannot be used for input and output at the same time.  So, in the
        # case of iterative plugins, if a dataset is used as output and then
        # as input on the next iteration, the subsequent output must be a
        # different file.
        # A cloned dataset is a copy of another dataset but with a different
        # backing file.  It doesn't have a name, is not accessible as a dataset
        # in the framework and is only used in alternation with another
        # dataset to allow it to be used as both input and output
        # simultaneously.
        
        # This is a cloned dataset (of out_dataset[0])
        self.create_clone(out_dataset[1], out_dataset[0])

        out_pData[0].plugin_data_setup('SINOGRAM', 'single')
        out_pData[1].plugin_data_setup('SINOGRAM', 'single')

        # input and output datasets for the first iteration
        self.set_iteration_datasets(0, [in_dataset[0]], [out_dataset[0]])
        # input and output datasets for subsequent iterations
        self.set_iteration_datasets(1, [in_dataset[0], out_dataset[0]],
                                    [out_dataset[1]])
        # out_dataset[0] and out_dataset[1] will continue to alternate for
        # all remaining iterations i.e. output becomes input and input becomes
        # output.

    # total number of output datasets
    def nOutput_datasets(self):
        return 2

    # total number of output datasets that are clones
    def nClone_datasets(self):
        return 1



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: replicate
   :platform: Unix
   :synopsis: A class to replicate the slice list of a dataset \
       (not the data itself!)

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import numpy as np
import copy

from savu.data.data_structures.data_types.base_type import BaseType


class Replicate(BaseType):
    """ Class to replicate the slice list of a dataset (not the data itself!)
    """

    def __init__(self, data_obj, reps):
        self.data_obj = data_obj
        self.reps = reps
        super(Replicate, self).__init__()

        self.rep_dim = len(data_obj.get_shape()) + 1
        self.shape = data_obj.get_shape() + (reps,)
        self.data = data_obj.data
        self.original_patterns = data_obj.get_data_patterns()
        self.__set_patterns(copy.deepcopy(self.original_patterns))

    def clone_data_args(self, args, kwargs, extras):
        args = ['self', 'reps']
        return args, kwargs, extras

    def __getitem__(self, idx):
        return np.expand_dims(self.data[idx[:-1]], self.rep_dim)

    def get_shape(self):
        return self.shape

    def __set_patterns(self, data_obj, patterns):
        for p in patterns:
            patterns[p]['slice_dir'] += (3,)
        data_obj.data_info.set('data_patterns', patterns)

    def _reset(self):
        self.data_obj.data_info.set('data_patterns', self.original_patterns)
        return self.data



# Copyright 2019 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: tomobar_recon_3D
   :platform: Unix
   :synopsis: A wrapper around TOmographic MOdel-BAsed Reconstruction (ToMoBAR) software \
   for advanced iterative image reconstruction using _3D_ capabilities of regularisation. \
   This plugin will divide 3D projection data into overalpped subsets using padding, which maskes it fast (GPU driver).

.. moduleauthor:: Daniil Kazantsev <scientificsoftware@diamond.ac.uk>
"""

from savu.plugins.reconstructions.base_recon import BaseRecon
from savu.data.plugin_list import CitationInformation
from savu.plugins.driver.gpu_plugin import GpuPlugin

import numpy as np
from tomobar.methodsIR import RecToolsIR
from savu.plugins.utils import register_plugin


@register_plugin
class TomobarRecon3d(BaseRecon, GpuPlugin):
    """
    A Plugin to reconstruct full-field tomographic projection data using state-of-the-art regularised iterative algorithms from \
    the ToMoBAR package. ToMoBAR includes FISTA and ADMM iterative methods and depends on the ASTRA toolbox and the CCPi RGL toolkit: \
    https://github.com/vais-ral/CCPi-Regularisation-Toolkit.

    :param output_size: The dimension of the reconstructed volume (only X-Y dimension). Default: 'auto'.
    :param padding: The amount of pixels to pad each slab of the cropped projection data. Default: 17.
    :param data_fidelity: Data fidelity, choose LS, PWLS, SWLS or KL. Default: 'LS'.
    :param data_Huber_thresh: Threshold parameter for __Huber__ data fidelity . Default: None.
    :param data_beta_SWLS: A parameter for stripe-weighted model. Default: 0.1.
    :param data_full_ring_GH: Regularisation variable for full constant ring removal (GH model). Default: None.
    :param data_full_ring_accelerator_GH: Acceleration constant for GH ring removal. Default: 10.0.
    :param algorithm_iterations: Number of outer iterations for FISTA (default) or ADMM methods. Default: 20.
    :param algorithm_verbose: print iterations number and other messages ('off' by default). Default: 'off'.
    :param algorithm_mask: set to 1.0 to enable a circular mask diameter or < 1.0 to shrink the mask. Default: 1.0.
    :param algorithm_ordersubsets: The number of ordered-subsets to accelerate reconstruction. Default: 6.
    :param algorithm_nonnegativity: ENABLE or DISABLE nonnegativity constraint. Default: 'ENABLE'.
    :param regularisation_method: To regularise choose methods ROF_TV, FGP_TV, PD_TV, SB_TV, LLT_ROF,\
                             NDF, TGV, NLTV, Diff4th. Default: 'FGP_TV'.
    :param regularisation_parameter: Regularisation (smoothing) value, higher \
                            the value stronger the smoothing effect. Default: 0.00001.
    :param regularisation_iterations: The number of regularisation iterations. Default: 80.
    :param regularisation_device: The number of regularisation iterations. Default: 'gpu'.
    :param regularisation_PD_lip: Primal-dual parameter for convergence. Default: 8.
    :param regularisation_methodTV:  0/1 - TV specific isotropic/anisotropic choice. Default: 0.
    :param regularisation_timestep: Time marching parameter, relevant for \
                    (ROF_TV, LLT_ROF, NDF, Diff4th) penalties. Default: 0.003.
    :param regularisation_edge_thresh: Edge (noise) related parameter, relevant for NDF and Diff4th. Default: 0.01.
    :param regularisation_parameter2:  Regularisation (smoothing) value for LLT_ROF method. Default: 0.005.
    :param regularisation_NDF_penalty: NDF specific penalty type Huber, Perona, Tukey. Default: 'Huber'.
    """

    def __init__(self):
        super(TomobarRecon3d, self).__init__("TomobarRecon3d")

    def _get_output_size(self, in_data):
        sizeX = self.parameters['output_size']
        shape = in_data.get_shape()
        if sizeX == 'auto':
            detX = in_data.get_data_dimension_by_axis_label('detector_x')
            sizeX = shape[detX]
        return sizeX

    def set_filter_padding(self, in_pData, out_pData):
        self.pad = self.parameters['padding']
        in_data = self.get_in_datasets()[0]
        det_y = in_data.get_data_dimension_by_axis_label('detector_y')
        pad_det_y = '%s.%s' % (det_y, self.pad)
        pad_dict = {'pad_directions': [pad_det_y], 'pad_mode': 'edge'}
        in_pData[0].padding = pad_dict
        out_pData[0].padding = pad_dict

    def setup(self):
        in_dataset = self.get_in_datasets()[0]
        self.parameters['vol_shape'] = self.parameters['output_size']            
        procs = self.exp.meta_data.get("processes")
        procs = len([i for i in procs if 'GPU' in i])
        dim = in_dataset.get_data_dimension_by_axis_label('detector_y')
        nSlices = int(np.ceil(in_dataset.get_shape()[dim]/float(procs)))
        self._set_max_frames(nSlices)      
        super(TomobarRecon3d, self).setup()

    def pre_process(self):
        in_pData = self.get_plugin_in_datasets()[0]
        out_pData = self.get_plugin_out_datasets()[0]
        detY = in_pData.get_data_dimension_by_axis_label('detector_y')
        # ! padding vertical detector !
        self.Vert_det = in_pData.get_shape()[detY] + 2*self.pad

        in_pData = self.get_plugin_in_datasets()
        self.det_dimX_ind = in_pData[0].get_data_dimension_by_axis_label('detector_x')
        self.det_dimY_ind = in_pData[0].get_data_dimension_by_axis_label('detector_y')
        self.output_size = out_pData.get_shape()[self.det_dimX_ind]
        
            # extract given parameters into dictionaries suitable for ToMoBAR input
        self._data_ = {'OS_number' : self.parameters['algorithm_ordersubsets'],
                       'huber_threshold' : self.parameters['data_Huber_thresh'],
                       'ringGH_lambda' :  self.parameters['data_full_ring_GH'],
                       'ringGH_accelerate' :  self.parameters['data_full_ring_accelerator_GH']}

        self._algorithm_ = {'iterations' : self.parameters['algorithm_iterations'],
			                'nonnegativity' : self.parameters['algorithm_nonnegativity'],
                            'mask_diameter' : self.parameters['algorithm_mask'],
                            'verbose' : self.parameters['algorithm_verbose']}

        self._regularisation_ = {'method' : self.parameters['regularisation_method'],
                                'regul_param' : self.parameters['regularisation_parameter'],
                                'iterations' : self.parameters['regularisation_iterations'],
                                'device_regulariser' : self.parameters['regularisation_device'],
                                'edge_threhsold' : self.parameters['regularisation_edge_thresh'],
                                'time_marching_step' : self.parameters['regularisation_timestep'],
                                'regul_param2' : self.parameters['regularisation_parameter2'],
                                'PD_LipschitzConstant' : self.parameters['regularisation_PD_lip'],
                                'NDF_penalty' : self.parameters['regularisation_NDF_penalty'],
                                'methodTV' : self.parameters['regularisation_methodTV']}

    def process_frames(self, data):
        cor, angles, self.vol_shape, init = self.get_frame_params()

        self.anglesRAD = np.deg2rad(angles.astype(np.float32))
        projdata3D = data[0].astype(np.float32)
        dim_tuple = np.shape(projdata3D)
        self.Horiz_det = dim_tuple[self.det_dimX_ind]
        half_det_width = 0.5*self.Horiz_det
        cor_astra = half_det_width - cor[0]
        projdata3D[projdata3D > 10**15] = 0.0
        projdata3D =np.swapaxes(projdata3D,0,1)
        #print(f"Shape of projdata3D is {np.shape(projdata3D)}")
        self._data_.update({'projection_norm_data' : projdata3D})

        # if one selects PWLS or SWLS models then raw data is also required (2 inputs)
        if ((self.parameters['data_fidelity'] == 'PWLS') or (self.parameters['data_fidelity'] == 'SWLS')):
            rawdata3D = data[1].astype(np.float32)
            rawdata3D[rawdata3D > 10**15] = 0.0
            rawdata3D = np.swapaxes(rawdata3D,0,1)/np.max(np.float32(rawdata3D))
            #print(f"Shape of rawdata3D is {np.shape(rawdata3D)}")
            self._data_.update({'projection_raw_data' : rawdata3D})
            self._data_.update({'beta_SWLS' : self.parameters['data_beta_SWLS']*np.ones(self.Horiz_det)})

       # set parameters and initiate a TomoBar class object
        self.Rectools = RecToolsIR(DetectorsDimH = self.Horiz_det,  # DetectorsDimH # detector dimension (horizontal)
                    DetectorsDimV = self.Vert_det,  # DetectorsDimV # detector dimension (vertical) for 3D case only
                    CenterRotOffset = cor_astra.item() - 0.5, # The center of rotation (CoR) scalar or a vector
                    AnglesVec = self.anglesRAD, # the vector of angles in radians
                    ObjSize = self.output_size, # a scalar to define the reconstructed object dimensions
                    datafidelity=self.parameters['data_fidelity'],# data fidelity, choose LS, PWLS, SWLS
                    device_projector='gpu')

        # Run FISTA reconstrucion algorithm here
        recon = self.Rectools.FISTA(self._data_, self._algorithm_, self._regularisation_)
        recon = np.swapaxes(recon, 0, 1)
        return recon

    def nInput_datasets(self):
        return max(len(self.parameters['in_datasets']), 1)

    def nOutput_datasets(self):
        return 1

    def _set_max_frames(self, frames):
        self._max_frames = frames

    def get_max_frames(self):
        return self._max_frames
    
    def get_slice_axis(self):
        return 'detector_y'

    def get_citation_information(self):
        cite_info1 = CitationInformation()
        cite_info1.name = 'citation1'
        cite_info1.description = \
            ("First-order optimisation algorithm for linear inverse problems.")
        cite_info1.bibtex = \
            ("@article{beck2009,\n" +
             "title={A fast iterative shrinkage-thresholding algorithm for linear inverse problems},\n" +
             "author={Amir and Beck, Mark and Teboulle},\n" +
             "journal={SIAM Journal on Imaging Sciences},\n" +
             "volume={2},\n" +
             "number={1},\n" +
             "pages={183--202},\n" +
             "year={2009},\n" +
             "publisher={SIAM}\n" +
             "}")
        cite_info1.endnote = \
            ("%0 Journal Article\n" +
             "%T A fast iterative shrinkage-thresholding algorithm for linear inverse problems\n" +
             "%A Beck, Amir\n" +
             "%A Teboulle, Mark\n" +
             "%J SIAM Journal on Imaging Sciences\n" +
             "%V 2\n" +
             "%N 1\n" +
             "%P 183--202\n" +
             "%@ --\n" +
             "%D 2009\n" +
             "%I SIAM\n")
        cite_info1.doi = "doi: "
        return cite_info1



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: mutations.py
   :platform: Unix
   :synopsis: A dictionary detailing changes to plugins, actions and \
       descriptions that are required by the configurator.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import textwrap
from colorama import Back, Fore
from . import display_formatter as df


def wrap(string):
    string = string.split('\n')
    final_str = []
    for s in string:
        final_str += textwrap.wrap(s, width=df.WIDTH)
    return '\n' + '\n'.join(final_str)


def header(string):
    return (Back.CYAN + string + Back.RESET + '\n')


def auto_replace_str():
    return header('***AUTO-REPLACEMENT NOTICE***:')


def auto_remove_str():
    return header('***AUTO-REMOVAL NOTICE***:')


def plugin_notice_str():
    return header('***PLUGIN NOTICE***:')


def param_changes_str(plugin):
    return header('***CHANGES TO PARAMETERS IN %s***:' % plugin)


def replace_str(old_name, new_name):
    return wrap(auto_replace_str() + '\n%s has been replaced by %s. '
                'Please check the parameters.\n' % (old_name, new_name))


def rename_str(old_name, new_name):
    return wrap(auto_replace_str() + '%s has been renamed as %s. No '
                'further action required.' % (old_name, new_name))


def remove_str(name, reason):
    return wrap(auto_remove_str() + '%s %s' % (name, reason))


def notice_str(name, notice):
    return wrap(plugin_notice_str() + '%s %s' % (name, notice))


def param_change_str(old, new, plugin, keys):
    removed = list(set(old).difference(set(new)))
    added = list(set(new).difference(set(old)))
    replaced = [entry['old'] for k in keys for entry in param_mutations[k]
                if entry['old'] in list(old.keys())]
    replacing = [entry['new'] for k in keys for entry in param_mutations[k]
                 if entry['old'] in list(old.keys())]

    removed = [x for x in removed if x not in replaced]
    added = [x for x in added if x not in replacing]

    if len(removed + added + replaced) > 0:
        removed_str = ["Removing parameter %s" % r for r in removed]
        added_str = ["Adding parameter %s" % a for a in added]
        replaced_str = ["Replacing parameter %s with %s" % (
                replaced[i], replacing[i]) for i in range(len(replaced))]
        print(wrap(param_changes_str(plugin) + '%s' % (
                '\n'.join(removed_str + added_str + replaced_str))))

hdf5_notice = 'is now used by default.\nPlease remove from the process list, '\
    'unless you wish to override the default parameters (which must be done '\
    'individually for each dataset).'

dezing_notice = '\nA faster and more accurate version of DezingFilter is now'\
    ' available as Dezinger.'

distortion_notice = 'A new version of DistortionCorrection is available with'\
    ' the version available in 2.3 \nand below being renamed as '\
    'DistortionCorrectionDeprecated.  Please replace with \nthe new version.'

medianfilt_notice = '\nThis version of MedianFilter is now deprecated. Please'\
' replace with the newer faster versions - MedianFilter or MedianFilterGpu.'

dezinger_dep_notice ='is now deprecated.  Please replace with the new version'\
                    ' of Dezinger (or DezingerGPU).'

dezinger_notice = '\nDezinger plugin has been replaced with a new version, '\
    'please update the parameters.  '\
    '\nNB: Dezinger should now be applied to normalised data. A GPU version '\
    'DezingerGPU is also available.'


plugin_mutations = \
    {'TimeseriesFieldCorrections':
        {'replace': 'DarkFlatFieldCorrection',
         'desc': replace_str('TimeseriesFieldCorrections',
                             'DarkFlatFieldCorrection')},
     'Hdf5TomoSaver':
        {'replace': 'Hdf5Saver',
         'desc': replace_str('Hdf5TomoSaver', 'Hdf5Saver')},
     'DezingFilter':
        {'replace': 'DezingerSimple',
         'desc': rename_str('DezingFilter', 'DezingerSimple') + wrap(dezing_notice)},
     'SavuLoader':
        {'replace': 'SavuNexusLoader',
         'desc': replace_str('SavuLoader', 'SavuNexusLoader')},
     'DistortionCorrection':
        {'replace': 'DistortionCorrectionDeprecated',
         'up_to_version': '2.4', # if the plist version is less than 2.4 (or not defined) then apply this mutation
         'desc': '\n' + Fore.RED + auto_replace_str() + wrap(distortion_notice) + Fore.RESET},
     'MedianFilter':
        {'replace': 'MedianFilterDeprecated',
         'up_to_version': '3.0', # if the plist version is less than 3.0 (or not defined) then apply this mutation
         'desc': '\n' + Fore.RED + auto_replace_str() + wrap(medianfilt_notice) + Fore.RESET},
     'Dezinger':
        {'replace': 'Dezinger',
         'up_to_version': '3.0', # if the plist version is less than 3.0 (or not defined) then apply this mutation
         'desc': '\n' + Fore.RED + auto_replace_str() + wrap(dezinger_notice) + Fore.RESET},
      'DezingerSimple':
         {'replace': 'DezingerSimpleDeprecated',
          'up_to_version': '3.0', # if the plist version is less than 3.0 (or not defined) then apply this mutation
          'desc': '\n' + Fore.RED + auto_replace_str() + wrap("\nDezingerSimple " + dezinger_dep_notice) + Fore.RESET},
      'DezingerSinogram':
         {'replace': 'DezingerSinogramDeprecated',
          'up_to_version': '3.0', # if the plist version is less than 3.0 (or not defined) then apply this mutation
          'desc': '\n' + Fore.RED + auto_replace_str() + wrap("\nDezingerSinogram " + dezinger_dep_notice) + Fore.RESET}        
     }

param_mutations = \
    {'BaseRecon': {'params': [{'old': 'center_of_rotation', 'new': 'centre_of_rotation'},
                   {'old': 'number_of_iterations', 'new': 'n_iterations'},
                   {'old': 'sino_pad', 'new': 'outer_pad'}],
                   'up_to_version': '2.4'},
     'AstraReconGpu': {'params': [{'old': 'reconstruction_type', 'new': 'algorithm'}],
                       'up_to_version': '2.4'},
     'AstraReconCpu': {'params': [{'old': 'reconstruction_type', 'new': 'algorithm'}],
                                  'up_to_version': '2.4'},
     'DistortionCorrection': {'params': [{'old': 'centre', 'new': 'centre_from_top', 'eval': 'val[0]'},
                              {'old': 'centre', 'new': 'centre_from_left', 'eval': 'val[1]'},
                              {'old': 'centre_x', 'new': 'centre_from_left'},
                              {'old': 'centre_y', 'new': 'centre_from_top'},
                              {'old': 'cod_from_left', 'new': 'centre_from_left'},
                              {'old': 'cod_from_top', 'new': 'centre_from_top'},
                              ], 'up_to_version': '2.4'}}

plugin_notices = \
    {'Hdf5Saver': {'desc': notice_str('Hdf5Saver', hdf5_notice)}, }



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: spectrum_crop
   :platform: Unix
   :synopsis: A plugin to crop a spectra

.. moduleauthor:: Aaron D. Parsons <scientificsoftware@diamond.ac.uk>
"""

import logging
from savu.plugins.filters.base_filter import BaseFilter
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin#,dawn_compatible
import numpy as np
import savu.test.test_utils as tu
from scipy.interpolate import griddata


@register_plugin
class ListToProjections(BaseFilter, CpuPlugin):
    """
    Converts a list of points into a 2D projection

    :param step_size_x: step size in the interp, None if minimum chosen. Default: None.
    :param step_size_y: step size in the interp, None if minimum chosen. Default: None.
    :param fill_value: the value to fill with, takes an average if nothing else chose. Default: 'mean'.

    """

    def __init__(self):
        logging.debug("interpolating projections")
        super(ListToProjections, self).__init__("ListToProjections")

    def pre_process(self):
        # assume all the projections are on the same axes
        in_datasets, _out_datasets = self.get_datasets()
        in_positions = in_datasets[0].meta_data.get("xy")

        if in_positions.ndim==3:
            positions = in_positions[0]# assume they are the same for all postiions
        elif in_positions.ndim==2:
            positions = in_positions# assume they are the same for all postiions
        self.setup_grids(positions)

    def process_frames(self, data):
        meshgridx, meshgridy = self.meshgrids
        data = data[0]
        if self.parameters['fill_value']=='mean':
            self.fill_value = data.mean()
        elif isinstance(self.parameters['fill_value'],float):
            self.fill_value = self.parameters['fill_value']
        elif isinstance(self.parameters['fill_value'],int):
            self.fill_value = float(self.parameters['fill_value'])
        else:
            logging.warning("I don't recognise your fill type of:%s , using 0 instead" % self.parameters['fill_value'])
            self.fill_value = 0

        return griddata((self.x, self.y),data, (meshgridx, meshgridy),fill_value=self.fill_value)[1:,1:]

    def setup(self):
        logging.debug('setting up the interpolation')
        in_dataset, out_datasets = self.get_datasets()
        in_pData, out_pData = self.get_plugin_datasets()

        inshape = in_dataset[0].get_shape()

        in_datasets, _out_datasets = self.get_datasets()
        in_positions = in_datasets[0].meta_data.get("xy")

        if in_positions.ndim==3:
            positions = in_positions[0]# assume they are the same for all postiions
        elif in_positions.ndim==2:
            positions = in_positions# assume they are the same for all postiions
        self.setup_grids(positions)
        out_projection_shape = self.meshgrids[0][1:,1:].shape
        in_pData[0].plugin_data_setup('PROJECTION', self.get_max_frames())
        proj_in_core_dirs= np.array(in_pData[0].get_core_dimensions())

        if len(proj_in_core_dirs)>1:
            raise IndexError("This plugin won't work since there are more than 1 core direction for the projection")

        outshape = list(inshape)
        proj_in_core_dirs = proj_in_core_dirs[0]
        outshape[proj_in_core_dirs] = out_projection_shape[0]
        outshape.insert(proj_in_core_dirs+1,out_projection_shape[1])

        axis_labels = in_datasets[0].get_axis_labels()
        axis_labels = [list(ix.keys())[0]+'.'+ix[list(ix.keys())[0]] for ix in axis_labels]
        axis_labels[proj_in_core_dirs] = 'x.microns'
        axis_labels.insert(proj_in_core_dirs+1,'y.microns')
        proj_out_core_dirs = (proj_in_core_dirs, proj_in_core_dirs+1)
        allDimsOut = list(range(len(outshape)))
        proj_out_slice_dirs = list(set(allDimsOut)-set(list(proj_out_core_dirs)))


        reshaped_projections = out_datasets[0]
        reshaped_projections.create_dataset(shape=tuple(outshape), axis_labels=axis_labels)
        reshaped_projections.add_pattern("PROJECTION", core_dims=proj_out_core_dirs, slice_dims=proj_out_slice_dirs)
        in_patterns = in_datasets[0].get_data_patterns()
        for pattern in list(in_patterns.keys()):
            if pattern != "PROJECTION":
                core_dir = in_patterns[pattern]['core_dims']
                slice_dir = list(set(allDimsOut)-set(core_dir))
                dim_info = {'core_dims': core_dir, 'slice_dims': slice_dir}
                reshaped_projections.add_pattern(pattern, **dim_info)
        out_pData[0].plugin_data_setup('PROJECTION', self.get_max_frames())
#
    def setup_grids(self,positions):
        #print positions.shape
        x = positions[0,:]
        y = positions[1,:]
        self.x = x
        self.y = y
        if self.parameters['step_size_x'] is not None:
            self.step_size_x = self.parameters['step_size_x']
        else:
            abs_diff_x = abs(np.diff(x))
            abs_diff_x_masked = abs_diff_x[abs_diff_x > 0.1]
            self.step_size_x = min(abs_diff_x_masked)

        if self.parameters['step_size_y'] is not None:
            self.step_size_y = self.parameters['step_size_y']
        else:
            abs_diff_y = abs(np.diff(y))
            abs_diff_y_masked = abs_diff_y[abs_diff_y > 0.1]
            self.step_size_y = min(abs_diff_y_masked)

        min_x = np.min(x)
        max_x = np.max(x)
        min_y = np.min(y)
        max_y = np.max(y)
        nptsx = ((max_x - min_x) / self.step_size_x) + 1
        nptsy = ((max_y - min_y) / self.step_size_y) + 1
        grid_x = np.arange(min_x, max_x, (max_x - min_x) / nptsx)
        grid_y = np.arange(min_y, max_y, (max_y - min_y) / nptsy)
        self.meshgrids = np.meshgrid(grid_x, grid_y)

    def get_max_frames(self):
        return 'single'

    def get_plugin_pattern(self):
        return 'PROJECTION'

    def nOutput_datasets(self):
        return 1



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: no_process_plugin
   :platform: Unix
   :synopsis: Plugin to test loading and saving without processing
.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""
from savu.plugins.plugin import Plugin
from savu.plugins.utils import register_plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin


@register_plugin
class NoProcessPlugin(Plugin, CpuPlugin):
    """
    The base class from which all plugins should inherit.
    :u*param pattern: Explicitly state the slicing pattern. Default: None.
    :param dummy: Dummy parameter for testing. Default: 10.
    """

    def __init__(self):
        super(NoProcessPlugin, self).__init__("NoProcessPlugin")

    def process_frames(self, data):
        return data[0]

    def setup(self):
        """
        Initial setup of all datasets required as input and output to the
        plugin.  This method is called before the process method in the plugin
        chain.
        """
        in_dataset, out_dataset = self.get_datasets()
        out_dataset[0].create_dataset(in_dataset[0])
        in_pData, out_pData = self.get_plugin_datasets()

        if self.parameters['pattern']:
            pattern = self.parameters['pattern']
        else:
            pattern = list(in_dataset[0].get_data_patterns().keys())[0]

        in_pData[0].plugin_data_setup(pattern, self.get_max_frames())
        out_pData[0].plugin_data_setup(pattern, self.get_max_frames())

    def get_max_frames(self):
        return 'multiple'

    def nInput_datasets(self):
        return 1

    def nOutput_datasets(self):
        return 1



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: tomo_recon
   :platform: Unix
   :synopsis: runner for tests using the MPI framework

.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""
import unittest
from savu.test import test_utils as tu
from savu.core.plugin_runner import PluginRunner


def run_protected_plugin_runner_no_process_list(options, plugin, **kwargs):
    if 'data' in list(kwargs.keys()):
        tu.set_plugin_list(options, plugin, kwargs['data'])
    else:
        tu.set_plugin_list(options, plugin)
    plugin_runner = PluginRunner(options)
    exp = plugin_runner._run_plugin_list()
    return exp


def run_protected_plugin_runner(options):
    plugin_runner = PluginRunner(options)
    exp = plugin_runner._run_plugin_list()
    return exp

if __name__ == "__main__":
    unittest.main()



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: yaml_loader
   :platform: Unix
   :synopsis: A class to load data from a non-standard nexus/hdf5 file using \
   descriptions loaded from a yaml file.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import os
import h5py
import yaml
import copy
import collections.abc as collections
import numpy as np  # used in exec so do not delete
from ast import literal_eval

import savu.plugins.utils as pu
import savu.plugins.loaders.utils.yaml_utils as yu
from savu.plugins.loaders.base_loader import BaseLoader
from savu.data.experiment_collection import Experiment


class YamlConverter(BaseLoader):
    """
    A class to load data from a non-standard nexus/hdf5 file using \
    descriptions loaded from a yaml file.

    :u*param yaml_file: Path to the file containing the data \
        descriptions. Default: None.
    :*param template_param: A hidden parameter to hold parameters passed in \
        via a savu template file. Default: {}.
    """

    def __init__(self, name='YamlConverter'):
        super(YamlConverter, self).__init__(name)

    def setup(self, template=False, metadata=True):
        #  Read YAML file
        yfile = self.parameters['yaml_file']
        data_dict = yu.read_yaml(self._get_yaml_file(yfile))
        data_dict = self._check_for_inheritance(data_dict, {})
        self._check_for_imports(data_dict)
        data_dict.pop('inherit', None)
        data_dict.pop('import', None)
        if template:
            return data_dict

        data_dict = self._add_template_updates(data_dict)
        self._set_entries(data_dict)

    def _get_yaml_file(self, yaml_file):
        if yaml_file is None:
            raise Exception('Please pass a yaml file to the yaml loader.')

        if not os.path.exists(yaml_file):
            path = os.path.dirname(
                __file__.split(os.path.join('savu', 'plugins'))[0])
            yaml_file = os.path.join(path, yaml_file)
            if not os.path.exists(yaml_file):
                raise Exception('The yaml file does not exist %s' % yaml_file)
        return yaml_file

    def _add_template_updates(self, ddict):
        all_entries = ddict.pop('all', {})
        for key, value in all_entries:
            for entry in ddict:
                if key in list(entry.keys()):
                    entry[key] = value

        for entry in self.parameters['template_param']:
            updates = self.parameters['template_param'][entry]
            ddict[entry]['params'].update(updates)
        return ddict

    def _check_for_imports(self, ddict):
        if 'import' in list(ddict.keys()):
            for imp in ddict['import']:
                name = False
                if len(imp.split()) > 1:
                    imp, name = imp.split('as')
                mod = __import__(imp.strip())
                globals()[mod.__name__ if not name else name] = mod

    def _check_for_inheritance(self, ddict, inherit, override=False):
        if 'inherit' in list(ddict.keys()):
            idict = ddict['inherit']
            idict = idict if isinstance(idict, list) else [idict]
            for i in idict:
                if i != 'None':
                    new_dict = yu.read_yaml(self._get_yaml_file(i))
                    new_dict, isoverride = \
                        self.__override(inherit, new_dict, override)
                    inherit.update(new_dict)
                    inherit = self._check_for_inheritance(
                            new_dict, inherit, override=isoverride)
        self._update(inherit, ddict)
        return inherit

    def __override(self, inherit, ddict, override):
        isoverride = False
        if 'override' in ddict:
            isoverride = ddict.pop('override')
        if override:
            for old, new in override.items():
                ddict[new] = ddict.pop(old)
                if new in list(inherit.keys()):
                    self._update(ddict[new], inherit[new])
        return ddict, isoverride

    def _update(self, d, u):
        for k, v in u.items():
            if isinstance(v, collections.Mapping):
                d[k] = self._update(d.get(k, {}), v)
            else:
                d[k] = v
        return d

    def _set_entries(self, ddict):
        entries = list(ddict.keys())
        for name in entries:
            self.get_description(ddict[name], name)

    def get_description(self, entry, name, metadata=True):
        # set params first as we may need them subsequently
        if 'params' in entry:
            self._set_params(entry['params'])
        # --------------- check for data entry -----------------------------
        if 'data' in list(entry.keys()):
            data_obj = self.exp.create_data_object("in_data", name)
            data_obj = self.set_data(data_obj, entry['data'])
            
        else:
            emsg = 'Please specify the data information in the yaml file.'
            raise Exception(emsg)

        if metadata:
            self._get_meta_data_descriptions(entry, data_obj)

    def _get_meta_data_descriptions(self, entry, data_obj):
        # --------------- check for axis label information -----------------
        if 'axis_labels' in list(entry.keys()):
            self._set_axis_labels(data_obj, entry['axis_labels'])
        else:
            raise Exception('Please specify the axis labels in the yaml file.')

        # --------------- check for data access patterns -------------------
        if 'patterns' in list(entry.keys()):
            self._set_patterns(data_obj, entry['patterns'])
        else:
            raise Exception('Please specify the patterns in the yaml file.')

        # add any additional metadata
        if 'metadata' in entry:
            self._set_metadata(data_obj, entry['metadata'])
        self.set_data_reduction_params(data_obj)

        if 'exp_metadata' in entry:
            self._set_metadata(data_obj, entry['exp_metadata'], exp=True)

    def set_data(self, name, entry):
        raise NotImplementedError('Please implement "set_data" function'
                                  ' in the loader')

    def _set_keywords(self, dObj):
        filepath = str(dObj.backing_file.filename)
        shape = str(dObj.get_shape())
        return {'dfile': filepath, 'dshape': shape}

    def __get_wildcard_values(self, dObj):
        if 'wildcard_values' in list(dObj.data_info.get_dictionary().keys()):
            return dObj.data_info.get('wildcard_values')
        return None

    def update_value(self, dObj, value, itr=0):
        import pdb
        # setting the keywords
        if dObj is not None:
            dshape = dObj.get_shape()
            dfile = dObj.backing_file
            globals()['dfile'] = dfile
            wildcard = self.__get_wildcard_values(dObj)

        if isinstance(value, str):
            split = value.split('$')
            if len(split) > 1:
                value = self._convert_string(dObj, split[1])
                try:
                    value = eval(value, globals(), locals())
                    value = self._convert_bytes(value)
                except Exception as e:
                    msg = (f"Error evaluating value: '{value}' \n %s" % e)
                    try:
                        value = value.replace("index(", "index(b")
                        value = eval(value, globals(), locals())
                        value = self._convert_bytes(value)
                    except:
                        value = eval(value, globals(), locals())
                        raise Exception(msg)
        return value

    def _convert_string(self, dObj, string):
        for old, new in self.parameters.items():
            if old in string:
                if isinstance(new, str):
                    split = new.split('$')
                    if len(split) > 1:
                        new = split[1]
                    elif isinstance(new, str): # nothing left to split
                        new = "'%s'" % new
                string = self._convert_string(
                        dObj, string.replace(old, str(new)))
        return string

    def _convert_bytes(self, value):
        # convert bytes to str - for back compatability
        if isinstance(value, bytes):
            return value.decode("ascii")
        if isinstance(value, np.ndarray) and isinstance(value[0], bytes):
            return value.astype(str)
        return value

    def _set_params(self, params):
        # Update variable parameters that are revealed in the template
        params = self._update_template_params(params)
        self.parameters.update(params)
        # find files, open and add to the namespace then delete file params
        files = [k for k in list(params.keys()) if k.endswith('file')]
        for f in files:
            param = params[f]
            try:
                globals()[str(f)] = self.update_value(None, param)
            except IOError:
                self._check_for_test_data(f, param)
            del params[f]

    def _check_for_test_data(self, f, param):
        # check if this is Savu test data
        substrs = param.split("'")[1:2]
        filename = None
        for s in substrs:
            try:
                filename = self._get_yaml_file(s)
                break
            except:
                pass
        param = param.replace(s, filename)
        globals()[str(f)] = self.update_value(None, param)
        del self.parameters[f]

    def _update_template_params(self, params):
        for k, v in params.items():
            v = pu.is_template_param(v)
            if v is not False:
                params[k] = \
                    self.parameters[k] if k in list(self.parameters.keys()) else v[1]
        return params

    def _set_axis_labels(self, dObj, labels):
        dims = list(range(len(list(labels.keys()))))
        axis_labels = [None]*len(list(labels.keys()))
        for d in dims:
            self._check_label_entry(labels[d])
            l = labels[d]
            for key in list(l.keys()):
                l[key] = self.update_value(dObj, l[key])
            axis_labels[l['dim']] = (l['name'] + '.' + l['units'])
            if l['value'] is not None:
                dObj.meta_data.set(l['name'], l['value'])
        dObj.set_axis_labels(*axis_labels)

    def _check_label_entry(self, label):
        required = ['dim', 'name', 'value', 'units']
        try:
            [label[i] for i in required]
        except:
            raise Exception("name, value and units are required fields for \
                            axis labels")

    def _set_patterns(self, dObj, patterns):
        for key, dims in patterns.items():
            core_dims = self.__get_tuple(
                    self.update_value(dObj, dims['core_dims']))
            slice_dims = self.__get_tuple(
                    self.update_value(dObj, dims['slice_dims']))
            dObj.add_pattern(key, core_dims=core_dims, slice_dims=slice_dims)

    def __get_tuple(self, val):
        return literal_eval(val) if not isinstance(val, tuple) else val

    def _set_metadata(self, dObj, mdata, exp=False):
        populate = dObj.exp if exp else dObj
        for key, value in mdata.items():
            value = self.update_value(dObj, value['value'])
            populate.meta_data.set(key, value)



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: temp_loader
   :platform: Unix
   :synopsis: A loader that creates a random number generated hdf5 dataset of\
       any size.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import os
import h5py
import logging
import numpy as np

from savu.data.chunking import Chunking
from savu.plugins.utils import register_plugin
from savu.plugins.loaders.base_loader import BaseLoader
from savu.plugins.savers.utils.hdf5_utils import Hdf5Utils


@register_plugin
class RandomHdf5Loader(BaseLoader):
    """
    A hdf5 dataset of a specified size is created at runtime using numpy\
    random sampling (numpy.random) and saved to file. This created dataset\
    will be used as the input file, and the input file path passed to Savu\
    will be ignored (use a dummy).

    :u*param size: A list specifiying the required data size. Default: [].
    :u*param axis_labels: A list of the axis labels to be associated with each\
    dimension, of the form ['name1.unit1', 'name2.unit2',...]. Default: [].
    :u*param patterns: A list of data access patterns e.g.\
    [SINOGRAM.0c.1s.2c, PROJECTION.0s.1c.2s], where 'c' and 's' represent core\
    and slice dimensions respectively and every dimension must be\
    specified.  Default: [].
    :u*param file_name: Assign a name to the created h5\
    file. Default: 'input_array'.
    :param dtype: A numpy array data type.  Default: 'int16'.
    :param dataset_name: The name assigned to the dataset. Default: 'tomo'.
    :param angles: A python statement to be evaluated or a file - if the value\
    is None, values will be in the interval [0, 180]. Default: None.
    :param pattern: Pattern used to create and store the hdf5 dataset - \
    default is the first pattern in the pattern dictionary. Default: None.
    :param range: Set the distribution interval. Default: [1, 10].
    """

    def __init__(self, name='RandomHdf5Loader'):
        super(RandomHdf5Loader, self).__init__(name)

    def setup(self):
        exp = self.exp
        data_obj = exp.create_data_object('in_data',
                                          self.parameters['dataset_name'])

        data_obj.set_axis_labels(*self.parameters['axis_labels'])
        self.__convert_patterns(data_obj)
        self.__parameter_checks(data_obj)

        data_obj.backing_file = self.__get_backing_file(data_obj)
        data_obj.data = data_obj.backing_file['/']['test']
        data_obj.data.dtype # Need to do something to .data to keep the file open!

        data_obj.set_shape(data_obj.data.shape)
        self.n_entries = data_obj.get_shape()[0]
        self._set_rotation_angles(data_obj, self._get_n_entries())
        return data_obj

    def __get_backing_file(self, data_obj):
        fname = '%s/%s.h5' % \
            (self.exp.get('out_path'), self.parameters['file_name'])

        if os.path.exists(fname):
            return h5py.File(fname, 'r')

        self.hdf5 = Hdf5Utils(self.exp)

        size = tuple(self.parameters['size'])

        patterns = data_obj.get_data_patterns()
        p_name = patterns[self.parameters['pattern']] if \
            self.parameters['pattern'] is not None else list(patterns.keys())[0]
        p_name = list(patterns.keys())[0]
        p_dict = patterns[p_name]
        p_dict['max_frames_transfer'] = 1
        nnext = {p_name: p_dict}

        pattern_idx = {'current': nnext, 'next': nnext}
        chunking = Chunking(self.exp, pattern_idx)
        chunks = chunking._calculate_chunking(size, np.int16)

        h5file = self.hdf5._open_backing_h5(fname, 'w')
        dset = h5file.create_dataset('test', size, chunks=chunks)

        self.exp._barrier()

        slice_dirs = list(nnext.values())[0]['slice_dims']
        nDims = len(dset.shape)
        total_frames = np.prod([dset.shape[i] for i in slice_dirs])
        sub_size = \
            [1 if i in slice_dirs else dset.shape[i] for i in range(nDims)]

        # need an mpi barrier after creating the file before populating it
        idx = 0
        sl, total_frames = \
            self.__get_start_slice_list(slice_dirs, dset.shape, total_frames)
        # calculate the first slice
        for i in range(total_frames):
            low, high = self.parameters['range']
            dset[tuple(sl)] = np.random.randint(
                low, high=high, size=sub_size, dtype=self.parameters['dtype'])
            if sl[slice_dirs[idx]].stop == dset.shape[slice_dirs[idx]]:
                idx += 1
                if idx == len(slice_dirs):
                    break
            tmp = sl[slice_dirs[idx]]
            sl[slice_dirs[idx]] = slice(tmp.start+1, tmp.stop+1)

        self.exp._barrier()

#        try:
#            h5file.close()
#        except:
#            logging.debug('There was a problem trying to close the file in random_hdf5_loader')

        return self.hdf5._open_backing_h5(fname, 'r')

    def __get_start_slice_list(self, slice_dirs, shape, n_frames):
        n_processes = len(self.exp.get('processes'))
        rank = self.exp.get('process')
        frames = np.array_split(np.arange(n_frames), n_processes)[rank]
        f_range = list(range(0, frames[0])) if len(frames) else []
        sl = [slice(0, 1) if i in slice_dirs else slice(None)
              for i in range(len(shape))]
        idx = 0
        for i in f_range:
            if sl[slice_dirs[idx]] == shape[slice_dirs[idx]]-1:
                idx += 1
            tmp = sl[slice_dirs[idx]]
            sl[slice_dirs[idx]] = slice(tmp.start+1, tmp.stop+1)

        return sl, len(frames)

    def __convert_patterns(self, data_obj):
        pattern_list = self.parameters['patterns']
        for p in pattern_list:
            p_split = p.split('.')
            name = p_split[0]
            dims = p_split[1:]
            core_dims = tuple([int(i[0]) for i in [d.split('c') for d in dims]
                              if len(i) == 2])
            slice_dims = tuple([int(i[0]) for i in [d.split('s') for d in dims]
                               if len(i) == 2])
            data_obj.add_pattern(
                    name, core_dims=core_dims, slice_dims=slice_dims)

    def _set_rotation_angles(self, data_obj, n_entries):
        angles = self.parameters['angles']

        if angles is None:
            angles = np.linspace(0, 180, n_entries)
        else:
            try:
                angles = eval(angles)
            except:
                raise Exception('Cannot set angles in loader.')

        n_angles = len(angles)
        data_angles = n_entries
        if data_angles != n_angles:
            raise Exception("The number of angles %s does not match the data "
                            "dimension length %s", n_angles, data_angles)
        data_obj.meta_data.set("rotation_angle", angles)

    def __parameter_checks(self, data_obj):
        if not self.parameters['size']:
            raise Exception(
                    'Please specifiy the size of the dataset to create.')

    def _get_n_entries(self):
        return self.n_entries



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: hdf5_transport_data
   :platform: Unix
   :synopsis: A data transport class that is inherited by Data class at \
   runtime. It organises the slice list and moves the data.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import os

from savu.data.transport_data.slice_lists import \
    SliceLists, GlobalData, LocalData
from savu.data.transport_data.base_transport_data import BaseTransportData


class Hdf5TransportData(BaseTransportData, SliceLists):
    """
    The Hdf5TransportData class performs the organising and movement of data.
    """

    def __init__(self, data_obj, name='Hdf5TransportData'):
        super(Hdf5TransportData, self).__init__(data_obj)
        self.mfp = None
        self.params = None
        if os.environ['savu_mode'] == 'basic':
            self.max_frames_function = self._calc_max_frames_transfer_single
        else:
            self.max_frames_function = self._calc_max_frames_transfer_multi

    def _get_slice_lists_per_process(self, dtype):
        pData = self.data._get_plugin_data()
        pData._set_padding_dict()
        self.pad = True if pData.padding else False
        self.transfer_data = GlobalData(dtype, self)
        trans_dict = self.transfer_data._get_dict()
        proc_dict = LocalData(dtype, self)._get_dict()
        return self.__combine_dicts(trans_dict, proc_dict)

    def __combine_dicts(self, d1, d2):
        for key, value in d2.items():
            d1[key] = value
        return d1

    def _get_padded_data(self, slice_list, end=False):
        return self.transfer_data._get_padded_data(slice_list, end=False)

    def _calc_max_frames_transfer(self, nFrames):
        return self.max_frames_function(nFrames)




# -*- coding: utf-8 -*-
# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
.. module:: refresh_process_lists_test
   :platform: Unix
   :synopsis: unittest test classes for plugins

.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""
import os
import unittest
import savu.test.test_process_list_utils as tplu
import scripts.config_generator.config_utils as cu
from scripts.config_generator.content import Content


class RefreshProcessListsTest(unittest.TestCase):

    def _refresh_process_file(self, path):
        content = Content()
        # open
        content.fopen(path, update=True)
        # refresh
        positions = content.get_positions()
        for pos_str in positions:
            content.refresh(pos_str)
        # save
        content.save(content.filename)


def generate_test(path):
    def test(self):
        self._refresh_process_file(path)
    return test


if __name__ == "__main__":
    cu.populate_plugins()
    path = os.path.dirname(os.path.realpath(__file__)).split('scripts')[0]

    nxs_in_tests, plugins_in_tests = \
        tplu.get_process_list(path + '/savu/test/travis')

    lists = tplu.get_test_process_list(path + 'test_data/process_lists') \
        + tplu.get_test_process_list(path+'test_data/test_process_lists')
    nxs_used = list(set(nxs_in_tests).intersection(set(lists)))

    test_path = path + '/test_data/test_process_lists'
    test_path2 = path + '/test_data/process_lists'
    exclude = ['multimodal/simple_fit_test_XRF.nxs']
    for f in [n for n in nxs_used if n not in exclude]:
        print("Refreshing process list", f, "...")
        if os.path.exists(os.path.join(test_path, f)):
            path = os.path.join(test_path, f)
        else:
            path = os.path.join(test_path2, f)

        # generates a function that calls refresh_process_file for each file
        test = generate_test(path)
        # then sets the function as an attribute of the test class, so that
        # every process list is ran as an independent test
        setattr(RefreshProcessListsTest, f"test_{path}", test)
    unittest.main()



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: I13SpeckleLoader
   :platform: Unix
   :synopsis: A class for loading I13's speckle tracking data

.. moduleauthor:: Aaron Parsons <scientificsoftware@diamond.ac.uk>

"""

from savu.plugins.loaders.base_loader import BaseLoader
from savu.plugins.utils import register_plugin
import h5py

@register_plugin
class I13SpeckleLoader(BaseLoader):
    """
    A class to load tomography data from an NXstxm file
    :param signal_key: Path to the signals.Default:'/entry/sample'.
    :param reference_key: Path to the reference.Default:'/entry/reference'.
    :param angle_key: Path to the reference.Default:'/entry/theta'.
    :param dataset_names: the output sets.Default: ['signal','reference'].
    """

    def __init__(self, name='I13SpeckleLoader'):
        super(I13SpeckleLoader, self).__init__(name)

    def setup(self):
        """

        """
        name_signal, name_reference = self.parameters['dataset_names']

        exp = self.exp
        signal = exp.create_data_object("in_data", name_signal)
        signal.backing_file = h5py.File(exp.meta_data.get("data_file"), 'r')

        reference = exp.create_data_object("in_data", name_reference)
        reference.backing_file = h5py.File(exp.meta_data.get("data_file"), 'r')

        signal.data = signal.backing_file[self.parameters['signal_key']]
        signal.set_shape(signal.data.shape)
        reference.data = signal.backing_file[self.parameters['reference_key']]
        reference.set_shape(signal.data.shape)

        signal.set_axis_labels('rotation_angle.degrees',
                               'idx.units',
                               'detector_y.pixel',
                               'detector_x.pixel')

        reference.set_axis_labels('rotation_angle.degrees',
                                  'idx.units',
                                  'detector_y.pixel',
                                  'detector_x.pixel')

        theta = signal.backing_file[self.parameters['angle_key']][0]
        signal.meta_data.set('rotation_angle', theta)

        ## hard code this stuff for now since it's probably going to change anyway
        signal.add_pattern("PROJECTION", core_dims=(-2,-1), slice_dims=(0,1))

        signal.add_pattern("SINOGRAM", core_dims=(0,-2), slice_dims=(1,-1))

        signal.add_pattern("4D_SCAN", core_dims=(1,-2,-1),
                        slice_dims=(0,))
  
        reference.add_pattern("PROJECTION", core_dims=(-2,-1), slice_dims=(0,1))

        reference.add_pattern("SINOGRAM", core_dims=(0,-1), slice_dims=(1,-1))

        reference.add_pattern("4D_SCAN", core_dims=(1,-2,-1),
                        slice_dims=(0,))
        self.set_data_reduction_params(signal)
        self.set_data_reduction_params(reference)



# Copyright 2015 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
.. module:: basic_transport
   :platform: Unix
   :synopsis: Transports data to and from file at the beginning and end of the\
   process list respectively.

.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""

import os

from savu.core.transport_setup import MPI_setup
from savu.plugins.savers.utils.hdf5_utils import Hdf5Utils
from savu.core.transports.hdf5_transport import Hdf5Transport
from savu.core.transports.base_transport import BaseTransport


class BasicTransport(BaseTransport):

    def __init__(self):
        super(BasicTransport, self).__init__()
        self.global_data = True
        self.h5trans = Hdf5Transport()
        self.data_flow = None
        self.count = 0
        self.hdf5_flag = True

    def _transport_initialise(self, options):
        MPI_setup(options)
        # initially reading from a hdf5 file so Hdf5TransportData will be used
        # for all datasets created in a loader
        options['transport'] = 'hdf5'
        os.environ['savu_mode'] = 'basic'

    def _transport_pre_plugin_list_run(self):
        # loaders have completed now revert back to BasicTransport, so any
        # output datasets created by a plugin will use this.
        self.hdf5 = Hdf5Utils(self.exp)
        self.data_flow = self.exp.meta_data.plugin_list._get_dataset_flow()
        plist = self.exp.meta_data.plugin_list
        self.n_plugins = plist._get_n_processing_plugins()
        self.final_dict = plist.plugin_list[-1]
        if self.n_plugins is not 1:
            self.exp.meta_data.set('transport', 'basic')

    def _transport_pre_plugin(self):
        trans = self.exp.meta_data.get('transport')
        if self.count == self.n_plugins - 1 or trans == 'hdf5':
            self.__set_hdf5_transport()

    def _transport_post_plugin(self):
        # revert back to basic if a temporary transport mechanism was used
        if self.hdf5_flag:
            self.__unset_hdf5_transport()

        if self.count == self.n_plugins - 2:
            self.exp.meta_data.set('transport', 'hdf5')

        if self.count == self.n_plugins - 1:  # final plugin
            self.h5trans.exp = self.exp
            self.h5trans.hdf5 = Hdf5Utils(self.exp)
            self.h5trans._transport_post_plugin()

        self.count += 1

    def __set_hdf5_transport(self):
        self.hdf5_flag = True
        self.exp.meta_data.set('transport', 'hdf5')
        files = self._get_filenames(self.final_dict)
        self._set_file_details(files)
        self._setup_h5_files()

    def __unset_hdf5_transport(self):
        self.exp.meta_data.set('transport', 'basic')
        self.hdf5_flag = False

    def _transport_terminate_dataset(self, data):
        if data.backing_file:
            self.hdf5._close_file(data)



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: hilbert_filter
   :platform: Unix
   :synopsis: A plugin to apply hilbert filter horizontally for tomo reconstruction of phase gradient images. Use before Vocentering. In AstraRecon set log to false and FBP_filter to none.

.. moduleauthor:: Tunhe Zhou <tunhe.zhou@diamond.ac.uk>

"""
import logging
import numpy as np
import pyfftw.interfaces.scipy_fftpack as fft


from savu.plugins.filters.base_filter import BaseFilter
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin


@register_plugin
class HilbertFilter(BaseFilter, CpuPlugin):
    """
    A plugin to apply Hilbert filter (for phase gradient image) on projections.

    """

    def __init__(self):
        """
        This is where the class is designed, and we inherit from 2 classes, the
        first is the Filter class, which deals with splitting the job up for
        us, and provides some simple methods which we need to overload. The
        second is the CpuPlugin class which tells the framework that the
        processing that you are doing here runs on 1 cpu.
        """
        logging.debug("initialising Hilbert filter")
        logging.debug("Calling super to make sure that all superclases are " +
                      " initialised")
        super(HilbertFilter, self).__init__('HilbertFilter')
        self.filter1 = None

    def pre_process(self):
        self._setup_hilbert(*self.get_plugin_in_datasets()[0].get_shape())

    def _setup_hilbert(self, height, width):
        centerx = np.ceil(width / 2.0)
        # Define the hilbert filter
        filter1 = np.ones((height, width), dtype=np.float32)
        filter1[:,0:int(centerx)] = filter1[:,0:int(centerx)]*(-1.0)
        self.filter1 = filter1

    def _hilbert(self, data):
        pci1 = fft.fft2(fft.fftshift(np.float32(data)))
        pci2 = fft.ifftshift(pci1)*self.filter1
        fpci0 = fft.ifftshift(fft.ifft2(fft.fftshift(pci2)))
        fpci = np.imag(fpci0)
        result = fpci
        return result

    def process_frames(self, data):
        proj = np.nan_to_num(data[0])
        return self._hilbert(proj)

    def get_max_frames(self):
        return 'single'




# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: base_fitter
   :platform: Unix
   :synopsis: a base class for all fitting methods

.. moduleauthor:: Aaron Parsons <scientificsoftware@diamond.ac.uk>

"""
import logging
from savu.plugins.plugin import Plugin
import numpy as np
import peakutils as pe
from savu.plugins.driver.cpu_plugin import CpuPlugin


class BaseFitter(Plugin, CpuPlugin):
    """
    This plugin fits peaks.
    :param in_datasets: Create a list of the dataset(s). Default: [].
    :param out_datasets: \
        A. Default: ["FitWeights", "FitWidths", "FitAreas", "residuals"].
    :param width_guess: An initial guess at the width. Default: 0.02.
    :param peak_shape: Which shape do you want. Default: "gaussian".
    :param PeakIndex: the peak index. Default: [].
    """

    def __init__(self, name='BaseFitter'):
        super(BaseFitter, self).__init__(name)

    def setup(self):
        # set up the output datasets that are created by the plugin
        in_dataset, out_datasets = self.get_datasets()

        shape = in_dataset[0].get_shape()
        outshape = tuple(shape[:-1])+(self.parameters('PeakIndex'),)
        axis_labels = ['-1.PeakIndex.pixel.unit']
        pattern_list = ['SINOGRAM', 'PROJECTION']

        fitAreas = out_datasets[0]
        fitHeights = out_datasets[1]
        fitWidths = out_datasets[2]

        fitAreas.create_dataset(patterns={in_dataset[0]: pattern_list},
                                axis_labels={in_dataset[0]: axis_labels},
                                shape=outshape)

        fitHeights.create_dataset(patterns={in_dataset[0]: pattern_list},
                                  axis_labels={in_dataset[0]: axis_labels},
                                  shape=outshape)

        fitWidths.create_dataset(patterns={in_dataset[0]: pattern_list},
                                 axis_labels={in_dataset[0]: axis_labels},
                                 shape=outshape)

        channel = {'core_dims': (-1,), 'slice_dims': list(range(len(shape)-1))}

        fitAreas.add_pattern("CHANNEL", **channel)
        fitHeights.add_pattern("CHANNEL", **channel)
        fitWidths.add_pattern("CHANNEL", **channel)
        #residlabels = in_dataset[0].meta_data.get('axis_labels')[0:3]
        #print residlabels.append(residlabels[-1])
        residuals = out_datasets[3]
        residuals.create_dataset(in_dataset[0])


        # setup plugin datasets
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('SPECTRUM', self.get_max_frames())

        out_pData[0].plugin_data_setup('CHANNEL', self.get_max_frames())
        out_pData[1].plugin_data_setup('CHANNEL', self.get_max_frames())
        out_pData[2].plugin_data_setup('CHANNEL', self.get_max_frames())
        out_pData[3].plugin_data_setup('SPECTRUM', self.get_max_frames())

    def get_max_frames(self):
        return 'single'

    def nOutput_datasets(self):
        return 4

    def setPositions(self, in_meta_data):
        try:
            positions = in_meta_data.get('PeakIndex')
            logging.debug('Using the pre-defined PeakIndex metadata')
        except KeyError:
            logging.error('No peaks defined!')

    def _resid(self, p, fun, y, x, pos):
        #print fun.__name__
        r = y-self._spectrum_sum(fun, x, pos, *p)

        return r

    def dfunc(self, p, fun, y, x, pos):
        if fun.__name__ == 'gaussian' or fun.__name__ == 'lorentzian':  # took the lorentzian out. Weird
            rest = p
            npts = len(p) // 2
            a = rest[:npts]
            sig = rest[npts:2*npts]
            mu = pos
            if fun.__name__ == 'gaussian':
                da = self.spectrum_sum_dfun(fun, 1. / a, x, mu, *p)
                dsig_mult = np.zeros((npts, len(x)))
                for i in range(npts):
                    dsig_mult[i] = ((x-mu[i])**2) / sig[i]**3
                dsig = self.spectrum_sum_dfun(fun, dsig_mult, x, mu, *p)
                op = np.concatenate([-da, -dsig])
            elif fun.__name__ == 'lorentzian':
                da = self.spectrum_sum_dfun(fun, 1. / a, x, mu, *p)
                dsig = np.zeros((npts, len(x)))
                for i in range(npts):
                    nom = 8 * a[i]* sig[i] * (x - mu[i]) ** 2
                    denom = (sig[i]**2 + 4.0 * (x - mu[i])**2)**2
                    dsig[i] = nom / denom
                op = np.concatenate([-da, -dsig])
        else:
            op = None
        return op

    def _spectrum_sum(self, fun, x, positions, *p):
        rest = np.abs(p)
        npts = len(p) // 2
        weights = rest[:npts]
        widths = rest[npts:2*npts]
        spec = np.zeros((len(x),))
        for ii in range(len(weights)):
            spec += fun(weights[ii], widths[ii], x, positions[ii])
        return spec

    def getFitFunction(self, key):
        self.lookup = {"lorentzian": lorentzian, "gaussian": gaussian}
        return self.lookup[key]

    def getFitFunctionNumArgs(self, key):
        self.lookup = {"lorentzian": 2, "gaussian": 2}
        return self.lookup[key]

    def getAreas(self, fun, x, positions, fitmatrix):
        rest = fitmatrix
        numargsinp = self.getFitFunctionNumArgs(str(fun.__name__))  # 2 in
        npts = len(fitmatrix) / numargsinp
        #print npts
        weights = rest[:npts]
        #print 'the weights are'+str(weights)
        widths = rest[npts:2*npts]
        #print 'the widths are'+str(widths)
        #print(len(widths))
        areas = []
        for ii in range(len(weights)):
            areas.append(np.sum(fun(weights[ii],
                                    widths[ii],
                                    x,
                                    positions[ii],
                                    )))
        return weights, widths, np.array(areas)

    def spectrum_sum_dfun(self, fun, multiplier, x, pos, *p):
        rest = p
        npts = len(p) // 2
        weights = rest[:npts]
        widths = rest[npts:2*npts]
        positions = pos
        spec = np.zeros((npts, len(x)))
        for ii in range(len(weights)):
            spec[ii] = multiplier[ii]*fun(weights[ii], widths[ii], x, positions[ii])
        return spec


def lorentzian(a, w, x, c):
    y = a / (1.0 + (2.0 * (c - x) / w) ** 2)
    return y


def gaussian(a, w, x, c):
    return pe.gaussian(x, a, c, w)



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: base_saver
   :platform: Unix
   :synopsis: A base class for all saver plugins

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

from savu.plugins.plugin import Plugin


class BaseSaver(Plugin):
    """
    A base plugin from which all data saver plugins should inherit.

    :*param out_datasets: Hidden, dummy out_datasets entry. Default: []
    :param in_datasets: The name of the dataset to save. Default: [].
    """

    def __init__(self, name="BaseSaver"):
        super(BaseSaver, self).__init__(name)
        self.frame = None

    def setup(self):
        in_pData = self.get_plugin_in_datasets()
        pattern = self.get_pattern()
        in_pData[0].plugin_data_setup(pattern, self.get_max_frames())

    def _get_group_name(self, name):
        nPlugin = self.exp.meta_data.get('nPlugin')
        plugin_dict = self.exp._get_collection()['plugin_dict'][nPlugin]
        return "%i-%s-%s" % (nPlugin, plugin_dict['name'], name)

    def get_pattern(self):
        return self.parameters['pattern']

    def get_frame(self):
        return self.frame

    def nInput_datasets(self):
        return 1

    def nOutput_datasets(self):
        return 0

    def get_max_frames(self):
        return 'single'



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: data_removal
   :platform: Unix
   :synopsis: Plugin to remove unwanted data.
.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import numpy as np

from savu.plugins.filters.base_filter import BaseFilter
from savu.plugins.utils import register_plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin


@register_plugin
class DataRemoval(BaseFilter, CpuPlugin):
    """
    A class to remove any unwanted data from the specified pattern frame.
    :param indices: A list or range of values to remove, e.g. [0, 1, 2], \
        0:2 (start:stop) or 0:2:1 (start:stop:step). Default: None.
    :param pattern: Explicitly state the slicing pattern. Default: 'SINOGRAM'.
    :param dim: Data dimension to reduce. Default: 0.
    """

    def __init__(self):
        super(DataRemoval, self).__init__("DataRemoval")
        self.indices = None
        self.sl = None

    def pre_process(self):
        in_pData = self.get_plugin_in_datasets()[0]
        in_data = self.get_in_datasets()[0]
        self.dim = self.parameters['dim']
        self.indices = self.calc_indices(in_data.get_shape()[self.dim])
        self.sl = [slice(None)]*len(in_pData.get_shape())
        self.sl[self.dim] = self.indices
        self.sl = tuple(self.sl)

        in_data.amend_axis_label_values(
            in_pData._get_data_slice_list(self.sl))

    def process_frames(self, data, frame_list):
        return data[0][self.sl]

    def setup(self):
        in_dataset, out_dataset = self.get_datasets()

        shape = list(in_dataset[0].get_shape())
        reduced_dim_shape = \
            len(self.calc_indices(shape[self.parameters['dim']]))
        shape[self.parameters['dim']] = reduced_dim_shape

        out_dataset[0].create_dataset(shape=tuple(shape),
                                      axis_labels=in_dataset[0],
                                      patterns=in_dataset[0])

        in_pData, out_pData = self.get_plugin_datasets()

        if self.parameters['pattern']:
            pattern = self.parameters['pattern']
        else:
            pattern = list(in_dataset[0].get_data_patterns().keys())[0]

        in_pData[0].plugin_data_setup(pattern, self.get_max_frames())
        out_pData[0].plugin_data_setup(pattern, self.get_max_frames())

    def calc_indices(self, orig_shape):
        indices = self.parameters['indices']
        if isinstance(indices, list):
            self.indices = np.array(indices)
        else:
            indices_list = indices.split(':')
            indices_list = [int(i) for i in indices_list]
            if len(indices_list) is 2:
                indices_list.append(1)
            self.indices = np.arange(*indices_list)
        return np.setxor1d(np.arange(orig_shape), self.indices)

    def nInput_datasets(self):
        return 1

    def nOutput_datasets(self):
        return 1

    def get_max_frames(self):
        return 'multiple'



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: data_additions
   :platform: Unix
   :synopsis: A module containing add_on classes, which have instances \
       encapsulated within the Data class.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import savu.data.data_structures.data_notes as notes
from savu.core.utils import docstring_parameter


class Padding(object):
    """ A class that organises padding of the data. An instance of Padding can
    be associated with a Data object in a plugin that inherits from BaseFilter,
    inside :meth:`savu.plugins.base_filter.BaseFilter.set_filter_padding`
    """

    def __init__(self, pData):
        self._pData = pData
        self.mtp = pData._get_max_frames_process()
        self.padding_dirs = {}
        self.pad_dict = None
        self.dims = None
        self.pattern_name = list(pData.get_pattern().keys())[0]
        self.pattern = pData.get_pattern()[self.pattern_name]
        self.dims = self.__set_dims()
        self.mode = 'edge'

    def __set_dims(self):
        dims = []
        core_dir = self.pattern['core_dims']
        slice_dir = self.pattern['slice_dims']
        for dim in range(len(core_dir + slice_dir)):
            self.padding_dirs[dim] = {'before': 0, 'after': 0}
        return dims

    def pad_mode(self, mode):
        self.mode = mode

    def pad_frame_edges(self, padding):
        """ Pad all the edges of a frame of data with the same pad amount
        (i.e pad in the core dimensions).

        :param int padding: The pad amount
        """
        core_dirs = self.pattern['core_dims']
        for core in core_dirs:
            self._pad_direction(str(core) + '.' + str(padding))

    def pad_multi_frames(self, padding):
        """ Add extra frames before and after the current frame of data (i.e
        pad in the fastest changing slice dimension).

        :param int padding: The pad amount
        """
        sdir = self.pattern['slice_dims'][0]
        self._pad_direction(str(sdir) + '.' + str(padding))

    @docstring_parameter(notes._padding.__doc__)
    def pad_directions(self, pad_list):
        """ Pad multiple, individually specified, dimensions.

        :param list(dict) pad_list: A list of strings of the form 'dim.pad',\
        'dim.after.pad' or 'dim.before.pad'
        {0}

        """
        for entry in pad_list:
            self._pad_direction(entry)

    @docstring_parameter(notes._padding.__doc__)
    def _pad_direction(self, pad_str):
        """ Pad the data in a specified dimension.

        :param str pad_str: A string of the form 'dim.pad', 'dim.after.pad'\
        or 'dim.before.pad'
        {0}

        """
        pad_vals = pad_str.split('.')
        pplace = None
        pad_place = ['before', 'after']
        if len(pad_vals) is 3:
            pdir, pplace, pval = pad_vals
            remove = list(set(pad_place).difference(set([pplace])))[0]
            pad_place.remove(remove)
        else:
            pdir, pval = pad_vals

        pdir = int(pdir)
        allowed_dims = \
            self.pattern['core_dims'] + (self.pattern['slice_dims'][0],)
        if pdir not in allowed_dims:
            raise Exception('Only core and first slice dimensions, %s, can be '
                            'padded. I cannot pad data dim %s in this plugin.'
                            % (allowed_dims, str(pdir)))
        else:
            for p in pad_place:
                self.padding_dirs[pdir][p] += int(pval)

    def _get_padding_directions(self):
        """ Get padding directions.

        :returns: padding dictionary
        :rtype: dict
        """
        for key in list(self.padding_dirs.keys()):
            if sum(self.padding_dirs[key].values()) is 0:
                del self.padding_dirs[key]
        return self.padding_dirs

    def _get_plugin_padding_directions(self):
        """ Get padding directions.

        :returns: padding dictionary
        :rtype: dict
        """
        for key in list(self.padding_dirs.keys()):
            if sum(self.padding_dirs[key].values()) is 0:
                del self.padding_dirs[key]
        return self.padding_dirs


class DataMapping(object):
    """ A class providing helper functions to multi-modal loaders.
    """

    def __init__(self):
        self._is_tomo = None
        self._is_map = None
        self._motors = None
        self._motor_type = None
        self._axes = None

    def set_motors(self, motors):
        self.motors = motors

    def get_motors(self):
        return self.motors

    def set_motor_type(self, motor_type):
        self.motor_type = motor_type

    def get_motor_type(self):
        return self.motor_type

    def set_axes(self, axes):
        self.axes = axes

    def get_axes(self):
        return self.axes

    def check_is_map(self, proj_dir):
        pattern = []
        if self._is_map:
            ovs = []
            for i in self.get_shape():
                if i != proj_dir[0]:
                    if i != proj_dir[1]:
                        ovs.append(i)
            pattern = \
                {"PROJECTION", {'core_dims': proj_dir, 'slice_dims': ovs}}
        return pattern

    def check_is_tomo(self, proj_dir, rotation):
        pattern = []
        if self._is_tomo:
            ovs = []
            for i in self.get_shape():
                if i != rotation:
                    if i != proj_dir[1]:
                        ovs.append(i)
            pattern = {"SINOGRAM", {'core_dims': (rotation, proj_dir[-1]),
                                    'slice_dims': ovs}}
        return pattern



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: Point-spread-function correction
   :platform: Unix
   :synopsis: A plugin for stitching two tomo-datasets.
.. moduleauthor:: Nghia Vo <scientificsoftware@diamond.ac.uk>

"""
import copy
import numpy as np
from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin
import sys


@register_plugin
class ImageStitching(Plugin, CpuPlugin):
    """
    Method to stitch images of two tomo-datasets including flat-field correction.
    :u*param overlap: Overlap width between two images. Default: 354.
    :u*param row_offset: Offset of row indices of projections in the second dataset \
    compared to the first dataset. Default: -1.
    :u*param crop: Parameters used to crop stitched image with respect to \
    the edges of an image. Format: [crop_top, crop_bottom, crop_left, crop_right]. Default: [0, 0, 250, 250].
    :u*param pattern: Data processing pattern is 'PROJECTION' or \
        'SINOGRAM'. Default: 'PROJECTION'.
    :param norm: Apply normalization before stitching. Default: True.
    :param flat_use: Apply flat-field correction. Default: True.
    """

    def __init__(self):
        super(ImageStitching, self).__init__('ImageStitching')

    def setup(self):
        in_dataset, out_dataset = self.get_datasets()
        in_pData, out_pData = self.get_plugin_datasets()
        self.space = self.parameters['pattern']
        
        if self.space == 'SINOGRAM':
            in_pData[0].plugin_data_setup('SINOGRAM_STACK', 2)
        else:
            in_pData[0].plugin_data_setup('PROJECTION_STACK', 2)
        rm_dim = in_dataset[0].get_slice_dimensions()[0]
        patterns = ['SINOGRAM.' + str(rm_dim), 'PROJECTION.' + str(rm_dim)]

        self.c_top, self.c_bot, self.c_left, self.c_right = self.parameters['crop']
        self.overlap = int(self.parameters['overlap'])
        self.row_offset = int(self.parameters['row_offset'])
        self.norm = self.parameters['norm']
        self.flat_use = self.parameters['flat_use']

        axis_labels = copy.copy(in_dataset[0].get_axis_labels())
        del axis_labels[rm_dim]
        width_dim = in_dataset[0].get_data_dimension_by_axis_label(
            'detector_x')
        height_dim = in_dataset[0].get_data_dimension_by_axis_label(
            'detector_y')
        shape = list(in_dataset[0].get_shape())

        shape[width_dim] = 2 * shape[width_dim] - \
            self.overlap - self.c_left - self.c_right
        if self.space == 'PROJECTION':
            shape[height_dim] = shape[height_dim] - self.c_top - self.c_bot
        del shape[rm_dim]
        out_dataset[0].create_dataset(
            patterns={in_dataset[0]: patterns},
            axis_labels=axis_labels,
            shape=tuple(shape))
        if self.space == 'SINOGRAM':
            out_pData[0].plugin_data_setup('SINOGRAM', 'single')
        else:
            out_pData[0].plugin_data_setup('PROJECTION', 'single')
        
    def make_weight_matrix(self, height1, width1, height2, width2, overlap, side):
        """
        Generate a linear-ramp weighting matrix for image stitching.

        Parameters
        ----------
        height1, width1 : int
            Size of the 1st image.
        height2, width2 : int
            Size of the 2nd image.
        overlap : int
            Width of the overlap area between two images.
        side : {0, 1}
            Only two options: 0 or 1. It is used to indicate the overlap side
            respects to image 1. "0" corresponds to the left side. "1" corresponds
            to the right side.
        """
        overlap = int(np.floor(overlap))
        wei_mat1 = np.ones((height1, width1), dtype=np.float32)
        wei_mat2 = np.ones((height2, width2), dtype=np.float32)
        if side == 1:
            list_down = np.linspace(1.0, 0.0, overlap)
            list_up = 1.0 - list_down
            wei_mat1[:, -overlap:] = np.float32(list_down)
            wei_mat2[:, :overlap] = np.float32(list_up)
        else:
            list_down = np.linspace(1.0, 0.0, overlap)
            list_up = 1.0 - list_down
            wei_mat2[:, -overlap:] = np.float32(list_down)
            wei_mat1[:, :overlap] = np.float32(list_up)
        return wei_mat1, wei_mat2

    def stitch_image(self, mat1, mat2, overlap, side, wei_mat1, wei_mat2, norm):
        """
        Stitch projection images or sinogram images using a linear ramp.

        Parameters
        ----------
        mat1 : array_like
            2D array. Projection image or sinogram image.
        mat2 :  array_like
            2D array. Projection image or sinogram image.
        overlap : float
            Width of the overlap area between two images.
        side : {0, 1}
            Only two options: 0 or 1. It is used to indicate the overlap side
            respects to image 1. "0" corresponds to the left side. "1" corresponds
            to the right side.
        wei_mat1 : array_like
            Weighting matrix used for image 1.
        wei_mat2 : array_like
            Weighting matrix used for image 2.
        norm : bool, optional
            Enable/disable normalization before stitching.

        Returns
        -------
        array_like
            Stitched image.
        """
        (nrow1, ncol1) = mat1.shape
        (nrow2, ncol2) = mat2.shape
        overlap_int = int(np.floor(overlap))
        sub_pixel = overlap - overlap_int
        if sub_pixel > 0.0:
            if side == 1:
                mat1 = shift(mat1, (0, sub_pixel), mode='nearest')
                mat2 = shift(mat2, (0, -sub_pixel), mode='nearest')
            else:
                mat1 = shift(mat1, (0, -sub_pixel), mode='nearest')
                mat2 = shift(mat2, (0, sub_pixel), mode='nearest')
        if nrow1 != nrow2:
            raise ValueError("Two images are not at the same height!!!")
        total_width0 = ncol1 + ncol2 - overlap_int
        mat_comb = np.zeros((nrow1, total_width0), dtype=np.float32)
        if side == 1:
            if norm is True:
                factor1 = np.mean(mat1[:, -overlap_int:])
                factor2 = np.mean(mat2[:, :overlap_int])
                mat2 = mat2 * factor1 / factor2
            mat_comb[:, 0:ncol1] = mat1 * wei_mat1
            mat_comb[:, (ncol1 - overlap_int):total_width0] += mat2 * wei_mat2
        else:
            if norm is True:
                factor2 = np.mean(mat2[:, -overlap_int:])
                factor1 = np.mean(mat1[:, :overlap_int])
                mat2 = mat2 * factor1 / factor2
            mat_comb[:, 0:ncol2] = mat2 * wei_mat2
            mat_comb[:, (ncol2 - overlap_int):total_width0] += mat1 * wei_mat1
        return mat_comb

    def pre_process(self):
        inData = self.get_in_datasets()[0]
        self.data_size = inData.get_shape()
        shape = len(self.get_plugin_in_datasets()[0].get_shape())
        (self.depth0, self.height0, self.width0, _) = inData.get_shape()
        self.sl1 = [slice(None)] * (shape - 1) + [0]
        self.sl2 = [slice(None)] * (shape - 1) + [1]
        if self.flat_use is True:
            dark = inData.data.dark_mean()
            flat = inData.data.flat_mean()
            (h_df, w_df) = dark.shape
            dark1 = dark[:h_df // 2]
            flat1 = flat[:h_df // 2]
            dark2 = dark[-h_df // 2:]
            flat2 = flat[-h_df // 2:]
            if self.space == 'SINOGRAM':            
                self.dark1 = 1.0 * dark1[:, self.c_left:]
                self.flat1 = 1.0 * flat1[:, self.c_left:]            
                self.dark2 = 1.0 * dark2[:, :w_df - self.c_right]
                self.flat2 = 1.0 * flat2[:, :w_df - self.c_right]            
            else:
                self.dark1 = 1.0 * \
                    dark1[self.c_top: h_df // 2 - self.c_bot, self.c_left:]
                self.flat1 = 1.0 * \
                    flat1[self.c_top: h_df // 2 - self.c_bot, self.c_left:]            
                dark2 = np.roll(dark2, self.row_offset, axis=0)
                flat2 = np.roll(flat2, self.row_offset, axis=0)
                self.dark2 = 1.0 * dark2[self.c_top: h_df // 2 -
                                         self.c_bot, :w_df - self.c_right]
                self.flat2 = 1.0 * flat2[self.c_top: h_df // 2 -
                                         self.c_bot, :w_df - self.c_right]            
            self.flatdark1 = self.flat1 - self.dark1
            self.flatdark2 = self.flat2 - self.dark2
            nmean = np.mean(np.abs(self.flatdark1))
            self.flatdark1[self.flatdark1 == 0.0] = nmean
            nmean = np.mean(np.abs(self.flatdark2))
            self.flatdark2[self.flatdark2 == 0.0] = nmean        
        if self.space == 'SINOGRAM':
            (_, w_df1) = self.dark1.shape
            (_, w_df2) = self.dark2.shape
            (self.wei1, self.wei2) = self.make_weight_matrix(
                self.depth0, w_df1, self.depth0, w_df2, self.overlap, 1)
        else:
            (h_df1, w_df1) = self.dark1.shape
            (h_df2, w_df2) = self.dark2.shape
            (self.wei1, self.wei2) = self.make_weight_matrix(
                h_df1, w_df1, h_df2, w_df2, self.overlap, 1)
            
        outData = self.get_out_datasets()[0]
        angles = inData.meta_data.get("rotation_angle")[:,0]
        outData.meta_data.set("rotation_angle", angles)

    def process_frames(self, data):
        mat1 = data[0][self.sl1]
        mat2 = data[0][self.sl2]
        if self.space == 'SINOGRAM':                       
            mat1 = np.float32(mat1[:, self.c_left:])
            mat2 = np.float32(mat2[:, :self.width0 - self.c_right])
            if self.flat_use is True:
                count = self.get_process_frames_counter()
                current_idx = self.get_global_frame_index()[count]
                mat1 = (mat1 - self.dark1[current_idx])\
                             / self.flatdark1[current_idx]
                mat2 = (mat2 - self.dark2[current_idx]) \
                            / self.flatdark2[current_idx]
        else:
            mat1 = np.float32(
                mat1[self.c_top:self.height0 - self.c_bot, self.c_left:])
            mat2 = np.roll(mat2, self.row_offset, axis=0)
            mat2 = np.float32(mat2[self.c_top:self.height0 -
                                   self.c_bot, :self.width0 - self.c_right])
            if self.flat_use is True:
                mat1 = (mat1 - self.dark1) / self.flatdark1
                mat2 = (mat2 - self.dark2) / self.flatdark2
        mat = self.stitch_image(mat1, mat2, self.overlap,
                                1, self.wei1, self.wei2, self.norm)
        return mat



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: reproduce_fit
   :platform: Unix
   :synopsis: A plugin to regenerate the fitted curves

.. moduleauthor:: Aaron Parsons <scientificsoftware@diamond.ac.uk>

"""
import logging
from savu.plugins.utils import register_plugin
from savu.plugins.fitters.base_fitter import BaseFitter
import numpy as np
from copy import deepcopy

@register_plugin
class ReproduceFit(BaseFitter):
    """
    This plugin reproduces the fitted curves. Its for diagnostics.
    :param in_datasets: Create a list of the dataset(s). Default: ["FitWeights", "FitWidths", "FitAreas","Backgrounds"].
    :param out_datasets: A. Default: ["Sum", "Individual_curves"].

    """

    def __init__(self):
        super(ReproduceFit, self).__init__("ReproduceFit")

    def pre_process(self):
        self.function = self.getFitFunction(self.parameters['peak_shape'])

    def process_frames(self, data):
        FitWeights,FitWidths, _FitAreas,Backgrounds = data
        npts = len(FitWeights)
        params = []
        params.extend(FitWeights)
        params.extend(FitWidths) 
        Sum = self._spectrum_sum(self.function, self.energy, self.PeakEnergy, *params) + Backgrounds
        Individual_curves = np.zeros((len(self.PeakEnergy), len(self.energy)))
        for i in range(npts):
            Individual_curves[i,:] = self.function(FitWeights[i], FitWidths[i], self.energy, self.PeakEnergy[i]) + Backgrounds
        return [Sum, Individual_curves]

    def setup(self):
        # set up the output datasets that are created by the plugin
        in_dataset, out_datasets = self.get_datasets()

        shape = in_dataset[-1].get_shape()

        summed = out_datasets[0]
        individual_curves = out_datasets[1]

        self.PeakEnergy = in_dataset[0].meta_data.get('PeakEnergy')
        self.energy = in_dataset[-1].meta_data.get('energy')

        new_shape = shape[:-1] + (len(self.PeakEnergy),) + (shape[-1],)
#         print "the new shape is"+str(new_shape)

        summed.create_dataset(in_dataset[-1])
        individual_curves.create_dataset(patterns=in_dataset[-1],
                                  axis_labels={in_dataset[-1]: ['~-2.idx.unit']},
                                  shape=new_shape)

        ss_slice = list(range(len(new_shape) - 2))
#         print "spectrum stack slice dims are" + str(ss_slice)
        spectrum_stack = {'core_dims': (-2, -1), 'slice_dims': tuple(ss_slice)}

        individual_curves.add_pattern("SPECTRUM_STACK", **spectrum_stack)

        # setup plugin datasets
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('CHANNEL', self.get_max_frames())
        in_pData[1].plugin_data_setup('CHANNEL', self.get_max_frames())
        in_pData[2].plugin_data_setup('CHANNEL', self.get_max_frames())
        in_pData[3].plugin_data_setup('SPECTRUM', self.get_max_frames())
        out_pData[0].plugin_data_setup('SPECTRUM', self.get_max_frames())
        out_pData[1].plugin_data_setup("SPECTRUM_STACK", self.get_max_frames())


    def nInput_datasets(self):
        return 4
    
    def nOutput_datasets(self):
        return 2



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: pyfai_azimuthal_integrator_with_bragg_filter
   :platform: Unix
   :synopsis: A plugin to integrate azimuthally "symmetric" signals i.e. SAXS,\
       WAXS or XRD.Requires a calibration file.

.. moduleauthor:: Aaron D. Parsons <scientificsoftware@diamond.ac.uk>
"""

import logging
import numpy as np
from savu.plugins.azimuthal_integrators.base_azimuthal_integrator \
    import BaseAzimuthalIntegrator
from savu.plugins.utils import register_plugin


@register_plugin
class PyfaiAzimuthalIntegratorWithBraggFilter(BaseAzimuthalIntegrator):
    """
    Uses pyfai to remap the data. We then remap, percentile file and integrate.

    :param use_mask: Should we mask. Default: False.
    :param num_bins: number of bins. Default: 1005.
    :param num_bins_azim: number of bins. Default: 200.
    :param thresh: threshold of percentile filter. Default: [5,95].
    """

    def __init__(self):
        logging.debug("Starting 1D azimuthal integration***")
        super(PyfaiAzimuthalIntegratorWithBraggFilter,
              self).__init__("PyfaiAzimuthalIntegratorWithBraggFilter")

    def process_frames(self, data):
        mData = self.params[2]
        ai = self.params[3]

        lims = self.parameters['thresh']
        num_bins_azim = self.parameters['num_bins_azim']
        num_bins_rad = self.parameters['num_bins']

        remapped, axis, _chi = \
            ai.integrate2d(data=data[0], npt_rad=num_bins_rad,
                           npt_azim=num_bins_azim, unit='q_A^-1')

        mask = np.ones_like(remapped)
        mask[remapped==0] = 0
        out = np.zeros(mask.shape[1])
        for i in range(mask.shape[1]):
#             print i
            idx = mask[:,i] == 1
            if np.sum(idx*1)==0:
                logging.warning("Found a bin where all the pixels are masked! Bin num: %s" , str(i))
                out[i] = 0.0
            else:
                foo = remapped[:,i][idx]
#                 print "the shape here is:"+str(foo.shape)
                top = np.percentile(foo,lims[1])
                bottom = np.percentile(foo,lims[0])
                out[i] = np.mean(np.clip(foo,bottom,top))
        return out



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: projection_vertical_alignment
   :platform: Unix
   :synopsis: Align each projection vertically with shift values calculated\
       using the ProjectionShift plugin

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import logging
import numpy as np
from scipy.ndimage.interpolation import shift as sci_shift

from savu.plugins.utils import register_plugin
from savu.plugins.filters.base_filter import BaseFilter
from savu.plugins.driver.cpu_plugin import CpuPlugin


@register_plugin
class ProjectionVerticalAlignment(BaseFilter, CpuPlugin):
    """
    Correct for vertical shift over projection images.

    :config_warn: Requires the PluginShift plugin to precede it.
    """

    def __init__(self):
        logging.debug("initialising Sinogram Alignment")
        super(ProjectionVerticalAlignment, self).__init__(
            "ProjectionVerticalAlignment")

    def pre_process(self):
        data = self.get_in_datasets()[0]
        self.shift = data.meta_data.get('proj_align_shift')
        self.slice_dir = data._get_plugin_data().get_slice_dimension()
        self.sl = [slice(None)]*3
        self.det_y = self.get_plugin_in_datasets()[0].\
            get_data_dimension_by_axis_label('detector_y')

    def process_frames(self, data):
        data = data[0]
        output = np.empty_like(data)
        nFrames = data.shape[self.slice_dir]
        sl = self.slice_list[self.slice_dir]
        entries = np.arange(sl.start, sl.stop, sl.step)

        for i in range(nFrames):
            self.sl[self.slice_dir] = i
            output[self.sl] = \
                sci_shift(data[self.sl], (self.shift[entries[i]][0], 0), cval=np.nan)

        return output



print("import: 'colorama'")
import colorama

print("import: 'h5py'")
import h5py

print("import: 'mpi4py'")
import mpi4py

print("import: 'numpy'")
import numpy

print("import: 'scipy'")
import scipy

print("import: 'pyfftw'")
import pyfftw

print("import: 'linecache2'")
import linecache2

print("import: 'traceback2'")
import traceback2

print("import: 'savu'")
import savu

print("import: 'yaml'")
import yaml




# Copyright 2019 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: morph_proc_line
   :platform: Unix
   :synopsis: a Larix module to remove inconsistent gaps in the resulted binary mask by merging the boundaries

.. moduleauthor:: Daniil Kazantsev <scientificsoftware@diamond.ac.uk>
"""

from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin

import numpy as np
from larix.methods.segmentation import MORPH_PROC_LINE

@register_plugin
class MorphProcLine(Plugin, CpuPlugin):
    """
    A Larix morphological processing module using line segments to remove inconsistent gaps

    :param primeclass: a mask class to start morphological processing from. Default: 0.
    :param correction_window: the size of the correction window. Default: 7.
    :param iterations: the number of iterations for segmentation. Default: 15.
    :param pattern: pattern to apply this to. Default: "VOLUME_YZ".
    """

    def __init__(self):
        super(MorphProcLine, self).__init__("MorphProcLine")

    def setup(self):
        in_dataset, out_dataset = self.get_datasets()
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup(self.parameters['pattern'], 'single')

        out_dataset[0].create_dataset(in_dataset[0], dtype=np.uint8)
        out_pData[0].plugin_data_setup(self.parameters['pattern'], 'single')

    def pre_process(self):
        # extract given parameters
        self.primeclass = self.parameters['primeclass']
        self.CorrectionWindow = self.parameters['correction_window']
        self.iterationsNumb = self.parameters['iterations']

    def process_frames(self, data):
        inputdata = data[0].copy(order='C')
        if (np.sum(inputdata) > 0):
            pars = {'maskdata' : np.uint8(inputdata),\
            'primeClass': self.primeclass,\
            'CorrectionWindow' : self.CorrectionWindow ,\
            'iterationsNumb' : self.iterationsNumb}
            # run class merging module here:
            mask_merged = MORPH_PROC_LINE(pars['maskdata'], pars['primeClass'],
                              pars['CorrectionWindow'], pars['iterationsNumb'])
        else:
            mask_merged = np.uint8(np.zeros(np.shape(inputdata)))
        return mask_merged

    def nInput_datasets(self):
        return 1
    def nOutput_datasets(self):
        return 1



# Copyright 2019 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: A plugin to initialise a binary mask for level sets and distance transform segmentations
   :platform: Unix
   :synopsis: A plugin to initialise a binary mask for level sets and distance transform segmentations

.. moduleauthor:: Daniil Kazantsev <scientificsoftware@diamond.ac.uk>
"""

from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin

import numpy as np

# using Morphological snakes module from
# https://github.com/pmneila/morphsnakes
from morphsnakes import circle_level_set

@register_plugin
class MaskInitialiser(Plugin, CpuPlugin):
    """
    A plugin to initialise a binary mask for level sets and distance transform segmentations.\
    Seeds are generated by providing coordinates of three points in 3D space (start-middle-finish) \
    and connecting them with a cylinder of a certain radius. \
    Importantly the Z coordinate is given following VOLUME_XY vertical pattern

    :param mask1_coordinates: X0,Y0,Z0 (start) X1,Y1,Z1 (middle) and X2,Y2,Z2 (finish) coordinates of three points. Default: [10, 10, 0, 15, 15, 15, 20, 20, 20].
    :param mask1_radius: Mask1 will be initialised with an ellipse of radius. Default: 5.
    :param mask2_coordinates: The second mask coordinates. Default: None.
    :param mask2_radius: Mask2 will be initialised with an ellipse of radius. Default: None.
    :param out_datasets: The default names . Default: ['INIT_MASK'].
    """

    def __init__(self):
        super(MaskInitialiser, self).__init__("MaskInitialiser")

    def setup(self):
        in_dataset, out_dataset = self.get_datasets()
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('VOLUME_YZ', 'single')

        out_dataset[0].create_dataset(in_dataset[0], dtype=np.uint8)
        out_pData[0].plugin_data_setup('VOLUME_YZ', 'single')
        self.full_input_shape = in_dataset[0].get_shape()

    def pre_process(self):
        # extract given parameters
        self.mask1_coordinates = self.parameters['mask1_coordinates']
        self.mask1_radius = self.parameters['mask1_radius']
        self.mask2_coordinates = self.parameters['mask2_coordinates']
        self.mask2_radius = self.parameters['mask2_radius']

        self.coordX = [0.0, 0.0, 0.0]
        self.coordY = [0.0, 0.0, 0.0]
        self.coordZ = [0.0, 0.0, 0.0]

        self.coordX[0] = self.mask1_coordinates[0]
        self.coordY[0] = self.mask1_coordinates[1]
        self.coordZ[0] = self.mask1_coordinates[2]
        self.coordX[1] = self.mask1_coordinates[3]
        self.coordY[1] = self.mask1_coordinates[4]
        self.coordZ[1] = self.mask1_coordinates[5]
        self.coordX[2] = self.mask1_coordinates[6]
        self.coordY[2] = self.mask1_coordinates[7]
        self.coordZ[2] = self.mask1_coordinates[8]

        if (self.mask2_coordinates is not None):
            self.coord2X = [0.0, 0.0, 0.0]
            self.coord2Y = [0.0, 0.0, 0.0]
            self.coord2Z = [0.0, 0.0, 0.0]

            self.coord2X[0] = self.mask2_coordinates[0]
            self.coord2Y[0] = self.mask2_coordinates[1]
            self.coord2Z[0] = self.mask2_coordinates[2]
            self.coord2X[1] = self.mask2_coordinates[3]
            self.coord2Y[1] = self.mask2_coordinates[4]
            self.coord2Z[1] = self.mask2_coordinates[5]
            self.coord2X[2] = self.mask2_coordinates[6]
            self.coord2Y[2] = self.mask2_coordinates[7]
            self.coord2Z[2] = self.mask2_coordinates[8]

    def process_frames(self, data):
        # get the index of a current frame
        index_current = self.get_plugin_in_datasets()[0].get_current_frame_idx()
        [dimX,dimY] = np.shape(data[0])
        mask = np.uint8(np.zeros(np.shape(data[0])))
        mask = mask_gen(mask, self.coordX, self.coordY, self.coordZ, self.mask1_radius, dimX, dimY, index_current)
        if ((self.mask2_coordinates is not None) and (self.mask2_radius is not None)):
            mask2 = np.uint8(np.zeros(np.shape(data[0])))
            mask2 = mask_gen(mask2, self.coord2X, self.coord2Y, self.coord2Z, self.mask2_radius, dimX, dimY, index_current)
            mask = np.add(mask, mask2)
        return mask

    def nInput_datasets(self):
        return 1
    def nOutput_datasets(self):
        return 1
    def get_max_frames(self):
        return 'single'


def mask_gen(mask, coordX, coordY, coordZ, mask_radius, dimX, dimY, index_current):
    steps1 = coordZ[1] - coordZ[0]
    distance1 = np.sqrt((coordX[1] - coordX[0])**2 + (coordY[1] - coordY[0])**2)
    d_dist1 = distance1/(steps1 - 1.0)
    d_step1 = d_dist1
    steps2 = coordZ[2] - coordZ[1]
    distance2 = np.sqrt((coordX[2] - coordX[1])**2 + (coordY[2] - coordY[1])**2)
    d_dist2 = distance2/(steps2 - 1.0)
    d_step2 = d_dist2
    if ((index_current >= coordZ[0]) & (index_current <= coordZ[1])):
        if ((coordX[0] == 0) & (coordY[0] == 0) & (coordX[1] == 0) & (coordY[1] == 0)):
            # create a full region mask (except the boundaries)
            bound_width=2 # outer boundary width
            mask[bound_width:dimX-bound_width, bound_width:dimY-bound_width] = 1
        else:
            d_step1 = (index_current - coordZ[0])*d_dist1
            if (distance1 != 0.0):
                t = d_step1/distance1
            else:
                t = 0.0
            if (coordX[0] == coordX[1]):
                x_t = np.int(coordX[0])
            else:
                x_t1 = np.round((1.0 - t)*coordX[0] + t*coordX[1])
                x_t = np.int(x_t1[0])
            if(coordY[0] == coordY[1]):
                y_t = np.int(coordY[0])
            else:
                y_t1 = np.round((1.0 - t)*coordY[0] + t*coordY[1])
                y_t = np.int(y_t1[0])
            mask = np.uint8(circle_level_set(np.shape(mask), (y_t, x_t), mask_radius))
    if ((index_current > coordZ[1]) & (index_current <= coordZ[2])):
        if ((coordX[1] == 0) & (coordY[1] == 0) & (coordX[2] == 0) & (coordY[2] == 0)):
            # create a full region mask (except the boundaries)
            bound_width=2 # outer boundary width
            mask[bound_width:dimX-bound_width, bound_width:dimY-bound_width] = 1
        else:
            d_step2 = (index_current - coordZ[1])*d_dist2
            if (distance2 != 0.0):
                t = d_step2/distance2
            else:
                t = 0.0
            if (coordX[1] == coordX[2]):
                x_t = np.int(coordX[1])
            else:
                x_t1 = np.round((1.0 - t)*coordX[1] + t*coordX[2])
                x_t = np.int(x_t1[0])
            if(coordY[1] == coordY[2]):
                y_t = np.int(coordY[1])
            else:
                y_t1 = np.round((1.0 - t)*coordY[1] + t*coordY[2])
                y_t = np.int(y_t1[0])
            mask = np.uint8(circle_level_set(np.shape(mask), (y_t, x_t), mask_radius))
    return [mask]



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: plugin_runner
   :platform: Unix
   :synopsis: Plugin list runner, which passes control to the transport layer.
.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>
"""

import logging

import savu.core.utils as cu
import savu.plugins.utils as pu
from savu.data.experiment_collection import Experiment


class BasicPluginRunner(object):
    """ Plugin list runner, which passes control to the transport layer.
    """

    def __init__(self, options, name='PluginRunner'):
        class_name = "savu.core.transports." + options["transport"] \
                     + "_transport"
        cu.add_base(self, cu.import_class(class_name))
        super(BasicPluginRunner, self).__init__()

        #  ********* transport function ***********
        self._transport_initialise(options)
        self.options = options
        # add all relevent locations to the path
        pu.get_plugins_paths()
        self.exp = Experiment(options)

    def _run_plugin_list(self):
        """ Create an experiment and run the plugin list.
        """
        self.exp.checkpoint = None
        plugin_list = self.exp.meta_data.plugin_list
        plugin_list._check_loaders()

        self.exp._set_nxs_filename()
        if self.exp.meta_data.get('process') == 0:
            fname = self.exp.meta_data.get('nxs_filename')
            plugin_list._save_plugin_list(fname)

        self.exp._set_loaders()

        #  ********* transport function ***********
        self._transport_pre_plugin_list_run()

        n_plugins = plugin_list._get_n_processing_plugins()
        n_loaders = self.exp.meta_data.plugin_list._get_n_loaders()

        plist = plugin_list.plugin_list
        count = 0
        for plugin_dict in plist[n_loaders:n_loaders+n_plugins]:
            self.exp.meta_data.set('nPlugin', count)
            self.__run_plugin(plugin_dict)
            count += 1

        # terminate any remaining datasets
        for data in list(self.exp.index['in_data'].values()):
            self._transport_terminate_dataset(data)

        cu.user_message("***********************")
        cu.user_message("* Processing Complete *")
        cu.user_message("***********************")

        logging.info('Processing complete')
        return self.exp

    def __run_plugin(self, plugin_dict):
        plugin = self._transport_load_plugin(self.exp, plugin_dict)
        self.exp.plugin = plugin
        plugin._main_setup(self.exp, plugin_dict['data'])

        self._transport_pre_plugin()

        plugin._run_plugin(self.exp, self)  # plugin driver
        self.exp._barrier()
        cu._output_summary(self.exp.meta_data.get("mpi"), plugin)
        plugin._clean_up()

        finalise = self.exp._finalise_experiment_for_current_plugin()

        #  ********* transport function ***********
        self._transport_post_plugin()

        for data in finalise['remove'] + finalise['replace']:
            #  ********* transport function ***********
            self._transport_terminate_dataset(data)

        self.exp._reorganise_datasets(finalise)



# -*- coding: utf-8 -*-
# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: experiment
   :platform: Unix
   :synopsis: Contains information specific to the entire experiment.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>
"""
from __future__ import print_function, division, absolute_import

import os
import copy
import h5py
import logging
from mpi4py import MPI

from savu.data.meta_data import MetaData
from savu.data.plugin_list import PluginList
from savu.data.data_structures.data import Data
from savu.core.checkpointing import Checkpointing
from savu.plugins.savers.utils.hdf5_utils import Hdf5Utils
import savu.plugins.loaders.utils.yaml_utils as yaml


class Experiment(object):
    """
    One instance of this class is created at the beginning of the
    processing chain and remains until the end.  It holds the current data
    object and a dictionary containing all metadata.
    """

    def __init__(self, options):
        self.meta_data = MetaData(options)
        self.__set_system_params()
        self.checkpoint = Checkpointing(self)
        self.__meta_data_setup(options["process_file"])
        self.collection = {}
        self.index = {"in_data": {}, "out_data": {}}
        self.initial_datasets = None
        self.plugin = None
        self._transport = None
        self._barrier_count = 0
        self._dataset_names_complete = False

    def get(self, entry):
        """ Get the meta data dictionary. """
        return self.meta_data.get(entry)

    def __meta_data_setup(self, process_file):
        self.meta_data.plugin_list = PluginList()
        try:
            rtype = self.meta_data.get('run_type')
            if rtype == 'test':
                self.meta_data.plugin_list.plugin_list = \
                    self.meta_data.get('plugin_list')
            else:
                raise Exception('the run_type is unknown in Experiment class')
        except KeyError:
            template = self.meta_data.get('template')
            self.meta_data.plugin_list._populate_plugin_list(process_file,
                                                             template=template)
        self.meta_data.set("nPlugin", 0) # initialise

    def create_data_object(self, dtype, name, override=True):
        """ Create a data object.

        Plugin developers should apply this method in loaders only.

        :params str dtype: either "in_data" or "out_data".
        """
        if name not in list(self.index[dtype].keys()) or override:
            self.index[dtype][name] = Data(name, self)
            data_obj = self.index[dtype][name]
            data_obj._set_transport_data(self.meta_data.get('transport'))
        return self.index[dtype][name]

    def _setup(self, transport):
        self._set_nxs_file()
        self._set_transport(transport)
        self.collection = {'plugin_dict': [], 'datasets': []}

        self._barrier()
        self._check_checkpoint()
        self._barrier()

    def _finalise_setup(self, plugin_list):
        checkpoint = self.meta_data.get('checkpoint')
        self._set_dataset_names_complete()
        # save the plugin list - one process, first time only
        if self.meta_data.get('process') == \
                len(self.meta_data.get('processes'))-1 and not checkpoint:
            # links the input data to the nexus file
            plugin_list._save_plugin_list(self.meta_data.get('nxs_filename'))
            self._add_input_data_to_nxs_file(self._get_transport())
        self._set_dataset_names_complete()

    def _set_initial_datasets(self):
        self.initial_datasets = copy.deepcopy(self.index['in_data'])

    def _update(self, plugin_dict):
        data = self.index['out_data'].copy()
        # clear output metadata after first setup
        for d in list(data.values()):
            d.meta_data._set_dictionary({})
        self.collection['datasets'].append(data)
        self.collection['plugin_dict'].append(plugin_dict)

    def _set_transport(self, transport):
        self._transport = transport

    def _get_transport(self):
        return self._transport

    def __set_system_params(self):
        sys_file = self.meta_data.get('system_params')
        import sys
        if sys_file is None:
            # look in conda environment to see which version is being used
            savu_path = sys.modules['savu'].__path__[0]
            sys_files = os.path.join(
                os.path.dirname(savu_path), 'system_files')
            subdirs = os.listdir(sys_files)
            sys_folder = 'dls' if len(subdirs) > 1 else subdirs[0]
            fname = 'system_parameters.yml'
            sys_file = os.path.join(sys_files, sys_folder, fname)
        logging.info('Using the system parameters file: %s', sys_file)
        self.meta_data.set('system_params', yaml.read_yaml(sys_file))

    def _check_checkpoint(self):
        # if checkpointing has been set but the nxs file doesn't contain an
        # entry then remove checkpointing (as the previous run didn't get far
        # enough to require it).
        if self.meta_data.get('checkpoint'):
            with h5py.File(self.meta_data.get('nxs_filename'), 'r') as f:
                if 'entry' not in f:
                    self.meta_data.set('checkpoint', None)

    def _add_input_data_to_nxs_file(self, transport):
        # save the loaded data to file
        h5 = Hdf5Utils(self)
        for name, data in self.index['in_data'].items():
            self.meta_data.set(['link_type', name], 'input_data')
            self.meta_data.set(['group_name', name], name)
            self.meta_data.set(['filename', name], data.backing_file)
            transport._populate_nexus_file(data)
            h5._link_datafile_to_nexus_file(data)

    def _set_dataset_names_complete(self):
        """ Missing in/out_datasets fields have been populated
        """
        self._dataset_names_complete = True

    def _get_dataset_names_complete(self):
        return self._dataset_names_complete

    def _reset_datasets(self):
        self.index['in_data'] = self.initial_datasets

    def _get_collection(self):
        return self.collection

    def _set_experiment_for_current_plugin(self, count):
        datasets_list = self.meta_data.plugin_list._get_datasets_list()[count:]
        exp_coll = self._get_collection()
        self.index['out_data'] = exp_coll['datasets'][count]
        if datasets_list:
            self._get_current_and_next_patterns(datasets_list)
        self.meta_data.set('nPlugin', count)

    def _get_current_and_next_patterns(self, datasets_lists):
        """ Get the current and next patterns associated with a dataset
        throughout the processing chain.
        """
        current_datasets = datasets_lists[0]
        patterns_list = {}
        for current_data in current_datasets['out_datasets']:
            current_name = current_data['name']
            current_pattern = current_data['pattern']
            next_pattern = self.__find_next_pattern(datasets_lists[1:],
                                                    current_name)
            patterns_list[current_name] = \
                {'current': current_pattern, 'next': next_pattern}
        self.meta_data.set('current_and_next', patterns_list)

    def __find_next_pattern(self, datasets_lists, current_name):
        next_pattern = []
        for next_data_list in datasets_lists:
            for next_data in next_data_list['in_datasets']:
                if next_data['name'] == current_name:
                    next_pattern = next_data['pattern']
                    return next_pattern
        return next_pattern

    def _set_nxs_file(self):
        folder = self.meta_data.get('out_path')
        fname = self.meta_data.get('datafile_name') + '_processed.nxs'
        filename = os.path.join(folder, fname)
        self.meta_data.set('nxs_filename', filename)

        if self.meta_data.get('process') == 1:
            if self.meta_data.get('bllog'):
                log_folder_name = self.meta_data.get('bllog')
                log_folder = open(log_folder_name, 'a')
                log_folder.write(os.path.abspath(filename) + '\n')
                log_folder.close()

        self._create_nxs_entry()

    def _create_nxs_entry(self):  # what if the file already exists?!
        logging.debug("Testing nexus file")
        import h5py
        if self.meta_data.get('process') == len(self.meta_data.get('processes')) - 1 and not self.checkpoint:
            with h5py.File(self.meta_data.get('nxs_filename'), 'w') as nxs_file:
                entry_group = nxs_file.create_group('entry')
                entry_group.attrs['NX_class'] = 'NXentry'

    def _clear_data_objects(self):
        self.index["out_data"] = {}
        self.index["in_data"] = {}

    def _merge_out_data_to_in(self):
        for key, data in self.index["out_data"].items():
            if data.remove is False:
                self.index['in_data'][key] = data
        self.index["out_data"] = {}

    def _finalise_experiment_for_current_plugin(self):
        finalise = {'remove': [], 'keep': []}
        # populate nexus file with out_dataset information and determine which
        # datasets to remove from the framework.

        for key, data in self.index['out_data'].items():
            if data.remove is True:
                finalise['remove'].append(data)
            else:
                finalise['keep'].append(data)

        # find in datasets to replace
        finalise['replace'] = []
        for out_name in list(self.index['out_data'].keys()):
            if out_name in list(self.index['in_data'].keys()):
                finalise['replace'].append(self.index['in_data'][out_name])

        
        return finalise

    def _reorganise_datasets(self, finalise):
        # unreplicate replicated in_datasets
        self.__unreplicate_data()

        # delete all datasets for removal
        for data in finalise['remove']:
            del self.index["out_data"][data.data_info.get('name')]

        # Add remaining output datasets to input datasets
        for name, data in self.index['out_data'].items():
            data.get_preview().set_preview([])
            self.index["in_data"][name] = copy.deepcopy(data)
        self.index['out_data'] = {}

    def __unreplicate_data(self):
        in_data_list = self.index['in_data']
        from savu.data.data_structures.data_types.replicate import Replicate
        for in_data in list(in_data_list.values()):
            if isinstance(in_data.data, Replicate):
                in_data.data = in_data.data._reset()

    def _set_all_datasets(self, name):
        data_names = []
        for key in list(self.index["in_data"].keys()):
            if 'itr_clone' not in key:
                data_names.append(key)
        return data_names

    def _barrier(self, communicator=MPI.COMM_WORLD, msg=''):
        comm_dict = {'comm': communicator}
        if self.meta_data.get('mpi') is True:
            logging.debug("Barrier %d: %d processes expected: %s",
                          self._barrier_count, communicator.size, msg)
            comm_dict['comm'].barrier()
        self._barrier_count += 1

    def log(self, log_tag, log_level=logging.DEBUG):
        """
        Log the contents of the experiment at the specified level
        """
        logging.log(log_level, "Experimental Parameters for %s", log_tag)
        for key, value in self.index["in_data"].items():
            logging.log(log_level, "in data (%s) shape = %s", key,
                        value.get_shape())
        for key, value in self.index["in_data"].items():
            logging.log(log_level, "out data (%s) shape = %s", key,
                        value.get_shape())



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: time_based_plus_drift_correction
   :platform: Unix
   :synopsis: A time-based dark and flat field correction that accounts for\
       image drift (Note: A work in progress but please try).

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import numpy as np
from skimage.feature import match_template
from scipy.ndimage.interpolation import shift as sci_shift

from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.corrections.time_based_correction import TimeBasedCorrection
from savu.plugins.utils import register_plugin


@register_plugin
class TimeBasedPlusDriftCorrection(TimeBasedCorrection, CpuPlugin):
    """
    Apply a time-based dark and flat field correction on data with an\
    image drift using linear interpolation and template matching.

    """

    def __init__(self):
        super(TimeBasedPlusDriftCorrection, self).__init__(
            "TimeBasedPlusDriftCorrection")

    def pre_process(self):
        super(TimeBasedPlusDriftCorrection, self).pre_process()

        self.shift_array = np.zeros((len(self.data_idx), 2))
        # find shift between flat field frames
        self.template = self.flat[0][100:300, 800:]
        #self.template = self.flat[0][10:20, 10:20]
        self.drift = self.calculate_flat_field_drift(self.template)

    def calculate_flat_field_drift(self, template):
        drift = []
        for i in range(len(self.flat)-1):
            drift.append(self.calculate_shift(
                self.flat[i], self.flat[i+1], template))
        return drift

    def calculate_shift(self, im1, im2, template):
        index = []
        for im in [im1, im2]:
            match = match_template(im, template)
            index.append(np.unravel_index(np.argmax(match), match.shape))
        index = np.array(index)
        shift = index[1] - index[0]
        return shift

    def calculate_flat_field(self, frame, data, frames, distance):
        shift = self.calculate_shift(self.flat[frames[0]], data, self.template)
        flat1 = sci_shift(self.flat[frames[0]], tuple(shift), cval=np.nan)
        flat2 = sci_shift(self.flat[frames[1]], shift-self.drift[frames[0]],
                          cval=np.nan)
        flat1, flat2 = self.fill_nans(flat1, flat2)

        if frames[0] > 0:
            shift = shift + self.drift[frames[0]-1]
        self.shift_array[np.where(self.data_idx == frame)[0]] = shift
        return flat1*distance[0] + flat2*distance[1]

    def fill_nans(self, im1, im2):
        im1[np.isnan(im1)] = im2[np.isnan(im1)]
        im2[np.isnan(im2)] = im1[np.isnan(im2)]
        return im1, im2

    def post_process(self):
        inData = self.get_in_datasets()[0]
        inData.meta_data.set('shift', self.shift_array)



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
.. module:: distortion_correction
   :platform: Unix
   :synopsis: A plugin to apply a distortion correction

.. moduleauthor:: Nghia Vo <scientificsoftware@diamond.ac.uk>

"""

import os
import logging
import numpy as np
from scipy.ndimage import map_coordinates
from scipy import interpolate

from savu.plugins.filters.base_filter import BaseFilter
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin
from savu.data.plugin_list import CitationInformation
import savu.core.utils as cu


@register_plugin
class DistortionCorrection(BaseFilter, CpuPlugin):
    """
    A plugin to apply radial distortion correction.

    :u*param polynomial_coeffs: Parameters of the radial distortion \
    function. Default: (1.0, 0.0e-1, 0.0e-2, 0.0e-3, 0.0e-4).
    :u*param center_from_top: The center of distortion in pixels from the top \
    of the image (referring to the original data size). Default: 1080.
    :u*param center_from_left: The center of distortion in pixels from the left \
    of the image (referring to the original data size). Default: 1280.
    :u*param file_path: Path to the text file having distortion coefficients\
     . Set to None for manually inputing. Default: None.
    :param crop_edges: When applied to previewed/cropped data, the result \
    may contain unwanted values around the edges, which can be removed by \
    cropping the edges by a specified number of pixels. Default: 0
    """

    def __init__(self):
        super(DistortionCorrection, self).__init__("DistortionCorrection")

    def load_metadata_txt(self, file_path):
        """
        Load distortion coefficients from a text file.
        Order of the infor in the text file:
        xcenter
        ycenter
        factor_0
        factor_1
        factor_2
        ..
        ---------
        Parameter:  - file_path: Path to the file
        ---------
        Return:     - Tuple of (xcenter, ycenter, list_fact).
        """
        with open(file_path, 'r') as f:
            x = f.read().splitlines()
            list_data = []
            for i in x:
                list_data.append(float(i.split()[-1]))
        xcenter = list_data[0]
        ycenter = list_data[1]
        list_fact = list_data[2:]
        return xcenter, ycenter, list_fact

    def setup(self):
        in_dataset, out_dataset = self.get_datasets()
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('PROJECTION', 'single')
        self.shape = list(in_dataset[0].get_shape())
        self.core_dims = in_pData[0].meta_data.get('core_dims')
        self.crop = self.parameters['crop_edges']
        for ddir in self.core_dims:
            self.shape[ddir] = self.shape[ddir] - 2 * self.crop
        out_dataset[0].create_dataset(patterns=in_dataset[0],
                                      axis_labels=in_dataset[0],
                                      shape=tuple(self.shape))
        out_pData[0].plugin_data_setup('PROJECTION', 'single')

    def pre_process(self):
        in_pData = self.get_plugin_in_datasets()[0]
        data = self.get_in_datasets()[0]
        name = data.get_name()
        shift = self.exp.meta_data.get(name + '_preview_starts')
        step = self.exp.meta_data.get(name + '_preview_steps')
        x_dim = data.get_data_dimension_by_axis_label('detector_x')
        y_dim = data.get_data_dimension_by_axis_label('detector_y')
        step_check = \
                True if max([step[i] for i in [x_dim, y_dim]]) > 1 else False
        if step_check:
            self.msg = "\n***********************************************\n"\
                "!!! ERROR !!! -> Plugin doesn't work with the step in the "\
                "preview larger than 1 \n"\
                "***********************************************\n"
            logging.warning(self.msg)
            cu.user_message(self.msg)
            raise ValueError(self.msg)

        x_offset = shift[x_dim]
        y_offset = shift[y_dim]
        file_path = self.parameters["file_path"]
        self.msg = ""
        x_center = 0.0
        y_center = 0.0
        if file_path is None:
            x_center = np.asarray(self.parameters['center_from_left'],
                                   dtype = np.float32) - x_offset
            y_center = np.asarray(self.parameters['center_from_top'],
                                   dtype = np.float32) - y_offset
            list_fact = np.float32(self.parameters['polynomial_coeffs'])
        else:
            if not (os.path.isfile(file_path)):
                self.msg = "!!! No such file: %s !!!"\
                        " Please check the file path" %str(file_path)
                cu.user_message(self.msg)
                raise ValueError(self.msg)
            try:
                (x_center, y_center, list_fact) = self.load_metadata_txt(
                    file_path)
                x_center = x_center - x_offset
                y_center = y_center - y_offset
            except IOError:
                self.msg = "\n*****************************************\n"\
                    "!!! ERROR !!! -> Can't open this file: %s \n"\
                    "*****************************************\n\
                    " % str(file_path)
                logging.warning(self.msg)
                cu.user_message(self.msg)
                raise ValueError(self.msg)

        data_shape = data.get_shape()
        self.height, self.width = data_shape[y_dim], data_shape[x_dim]
        xu_list = np.arange(self.width) - x_center
        yu_list = np.arange(self.height) - y_center
        xu_mat, yu_mat = np.meshgrid(xu_list, yu_list)
        ru_mat = np.sqrt(xu_mat**2 + yu_mat**2)
        fact_mat = np.sum(
            np.asarray([factor * ru_mat**i for i,
                        factor in enumerate(list_fact)]), axis=0)
        xd_mat = np.float32(np.clip(
            x_center + fact_mat * xu_mat, 0, self.width - 1))
        yd_mat = np.float32(np.clip(
            y_center + fact_mat * yu_mat, 0, self.height - 1))

        diff_y = np.max(yd_mat) - np.min(yd_mat)
        if (diff_y<1):
            self.msg = "\n*****************************************\n\n"\
                    "!!! ERROR !!! -> You need to increase the preview size"\
                    " for this plugin to work \n\n"\
                    "*****************************************\n"
            logging.warning(self.msg)
            cu.user_message(self.msg)

            raise ValueError(self.msg)
        self.indices = np.reshape(yd_mat, (-1, 1)),\
                        np.reshape(xd_mat, (-1, 1))

    def process_frames(self, data):
        mat_corrected = np.reshape(map_coordinates(
                data[0], self.indices, order=1, mode='reflect'),
             (self.height, self.width))
        return mat_corrected[self.crop:self.height - self.crop,
                                       self.crop:self.width - self.crop]

    def executive_summary(self):
        if self.msg != "":
            cu.user_message(self.msg)
            raise ValueError(self.msg)
        else:
            return ["Nothing to Report"]

    def get_citation_information(self):
        cite_info = CitationInformation()
        cite_info.description = \
            ("The distortion correction used in this processing chain is taken\
             from this work.")
        cite_info.bibtex = \
            ("@article{Vo:15,\n" +
             "title={Radial lens distortion correction with sub-pixel accuracy \
             for X-ray micro-tomography},\n" +
             "author={Nghia T. Vo and Robert C. Atwood and \
             Michael Drakopoulos},\n" +
             "journal={Optics. Express},\n" +
             "volume={23},\n" +
             "number={25},\n" +
             "pages={32859--32868},\n" +
             "year={2015},\n" +
             "publisher={OSA Publishing}" +
             "}")
        cite_info.doi = "doi: DOI: 10.1364/OE.23.032859"
        return cite_info
